{***************************************************************************************************
 This file is part of the GLOSSA Programming Language Interpreter (  )
 Copyright (C) 2002-2008 Alkis Georgopoulos - http://users.sch.gr/alkisg/

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************************************}
unit uEkfraseis;
{***************************************************************************************************
      .
***************************************************************************************************}
interface

uses
  Variants, SysUtils, Math, uLektikosAvalutns, uSustatika, uRu8miseis, uLa8os;

const
  KevoStr = '';
type
  TSuv8etnEkfrasn = class(TSuv8etoSustatiko)
{***************************************************************************************************
          .
***************************************************************************************************}
  protected
    orismata: TListaSustatikwv;
    function    Leitourgies: Word; override;  //lsDesTimn [or lsSta8ernTimn] [or lsSta8ernApotimnsn]
  public
    constructor Create;                     //    ,   orismata
    function    Perigrafn: WideString; override;               //  ,    
    destructor  Destroy; override;       //    override,   orismata
  end;

  TDeiktes = array of integer;

  TEkfrasnSeParev8esn = class(TSuv8etnEkfrasn)                             //orismata[0] =  
  protected                    //   Leitourgies ,  ..  ()  
    function    DesTimn: Variant; override;
    function    DesOrisma: TSustatiko;
  public
    constructor Create(dk: TDiaxeiristnsKwdika);
    function    TuposDedomevwv: TVarType; override;
    property    Orisma: TSustatiko read desOrisma;         //     
  end;

  TSuvSarEkfrasns = function(dk: TDiaxeiristnsKwdika): TSustatiko;

  TMovadiaiaPrajn = class(TSuv8etnEkfrasn)                                 //orismata[0] =  
  protected                 //  -, +,        
    prajn: TAploSustatiko;
    fTuposDedomevwv: TVarType;
    function    DesTimn: Variant; override;
  public
    constructor Create(dk: TDiaxeiristnsKwdika; sse: TSuvSarEkfrasns);
    function    TuposDedomevwv: TVarType; override;
  end;

  TDuadiknPrajn = class(TSuv8etnEkfrasn)                              //orismata[0, 1] =  
  protected
    prajn: TAploSustatiko;
    fTuposDedomevwv: TVarType;
    function    DesTimn: Variant; override;
  public
    constructor Create(dk: TDiaxeiristnsKwdika; orisma1: TSustatiko; sse: TSuvSarEkfrasns);
// TDuadiknPrajn.Create      ,     
//              .
    function    TuposDedomevwv: TVarType; override;
  end;

  function Ekfrasn(dk: TDiaxeiristnsKwdika): TSustatiko;
  function VariantWsKeimevo(const v: Variant): WideString;

{**************************************************************************************************}
implementation
{**************************************************************************************************}

{ TSuv8etnEkfrasn }

function TSuv8etnEkfrasn.Leitourgies: Word;
{***************************************************************************************************
    inherited Leitourgies        
    ,   lsSta8ernTimn   .
   ()        .
***************************************************************************************************}
//todo:    !  1*2*...*25!
//      ,   ""    ...
var
  i: integer;
  st, sa: boolean;
begin
  result := lsDesTimn;
  st := true;
  sa := true;
  for i := 0 to orismata.Pln8os - 1 do
  begin
    if not orismata[i].Ypostnrizei(lsSta8ernTimn) then
      st := false;
    if not orismata[i].Ypostnrizei(lsSta8ernApotimnsn) then
      sa := false;
  end;
  if st then
    result := result + lsSta8ernTimn;
  if sa then
    result := result + lsSta8ernApotimnsn;
end;

constructor TSuv8etnEkfrasn.Create;
begin
  orismata := TListaSustatikwv.Create;
end;

function TSuv8etnEkfrasn.Perigrafn: WideString;
begin
  case TuposDedomevwv and not VarArray of      // TuposDedomevwv  virtual,   
    varInt64:   result := ' ';
    varDouble:  result := ' ';
    varOleStr:  result := ' ';
    varBoolean: result := ' ';
  else
    Assert(true);
  end;
  if TuposDedomevwv and VarArray <> 0 then
    result := result + ' - ';  
end;

destructor TSuv8etnEkfrasn.Destroy;
begin
  FreeAndNil(orismata);
  inherited;
end;

{ TEkfrasnSeParev8esn }

function TEkfrasnSeParev8esn.DesTimn: Variant;
begin
  result := orismata[0].Timn;
end;

function TEkfrasnSeParev8esn.DesOrisma: TSustatiko;
{***************************************************************************************************
      -    ,    
    .
***************************************************************************************************}
begin
  if orismata[0] is TEkfrasnSeParev8esn then
    result := (orismata[0] as TEkfrasnSeParev8esn).DesOrisma       //   
  else result := orismata[0];  
end;

constructor TEkfrasnSeParev8esn.Create(dk: TDiaxeiristnsKwdika);
{***************************************************************************************************
       .
***************************************************************************************************}
resourcestring
  sAvoixtnParev8esn = '     .' + WBR + '%s';
begin
  inherited Create;                                              //    
  fArxiko := dk.TrSus;
  dk.Epomevo;
  orismata.Eisagwgn(Ekfrasn(dk));
  fTeliko := dk.TrSus;
  if fTeliko.Tupos <> sDeParev8esn then
    raise ESuvtaktikoLa8os.CreateFmt(sAvoixtnParev8esn, [orismata[0].KwdikasKaiPerigrafn],
      orismata[0].Perioxn);
  dk.Epomevo;                 //           
end;

function TEkfrasnSeParev8esn.TuposDedomevwv: TVarType;
begin
  result := orismata[0].TuposDedomevwv;
end;

{ TMovadiaiaPrajn }

function TMovadiaiaPrajn.DesTimn: Variant;
{***************************************************************************************************
              
 .      .
***************************************************************************************************}
resourcestring
  sSfalmaApotimnsnsEvswmatwmevnsSuvartnsns =
   ' %s       %s.' + WBR
   + ' : %s = %s';
var
  timn: Variant;
begin
  timn := orismata[0].Timn;                 //        
  try
    case prajn.Tupos of
      sMovadiaioMeiov, sDuadikoPlnv: result := - timn;
      sMovadiaioSuv, sDuadikoSuv:    result := + timn;
      sOXI:                          result := not timn;
      sA_M:                          if ru8miseis[ruMa8nmatikoAM].timn then
                                       result := Floor(timn)
                                     else
                                       result := Trunc(timn);
      sA_T:                          result := Abs(timn);    // int64,    VarAsType
      sE:                            result := Exp(extended(timn));   // D2006   timn
      sEF:                           if ru8miseis[ruParametrosHMSeMoires].timn then    //  single!
                                       result := Tan(Pi*extended(timn)/180)
                                     else
                                       result := Tan(extended(timn));
      sHM:                           if ru8miseis[ruParametrosHMSeMoires].timn then
                                       result := Sin(Pi*extended(timn)/180)
                                     else
                                       result := Sin(extended(timn));
      sLOG:                          result := Ln(extended(timn));
      sSYNnm:                        if ru8miseis[ruParametrosHMSeMoires].timn then
                                       result := Cos(Pi*extended(timn)/180)
                                     else
                                       result := Cos(extended(timn));
      sT_R:                          result := Sqrt(extended(timn));
    else
      Assert(true);
    end;
    result := VarAsType(result, fTuposDedomevwv);  //   int64   integer
  except
    on E: exception do      //todo:   reraise,       
      raise ELa8osEktelesns.CreateFmt(sSfalmaApotimnsnsEvswmatwmevnsSuvartnsns,
        [E.message, prajn.Kwdikas, orismata[0].Kwdikas, VariantWsKeimevo(timn)], Perioxn);
  end;
end;

constructor TMovadiaiaPrajn.Create(dk: TDiaxeiristnsKwdika; sse: TSuvSarEkfrasns);
{***************************************************************************************************
    ,   +, -   .
   TrSus =  .
***************************************************************************************************}
resourcestring
  sDevAvoijeParev8esn = ' (     %s.';
  sDevEkleiseParev8esn = '        %s.';
  sAsumbatoOrismaSuvartnsns = '   %s       .'
    + WBR + '%s';
  sAsumbatoOrismaTelestn = '  %s       .' + WBR + '%s';
var
  suvartnsn: boolean;
begin
  inherited Create;                                              //    
  fArxiko := dk.TrSus;
  prajn := fArxiko;
  dk.Epomevo;                                                                 //  
  suvartnsn := prajn.Tupos in [sA_M, sA_T, sE, sEF, sHM, sLOG, sSYNnm, sT_R];
  if suvartnsn then                                              // ,   sse
  begin
    if dk.TrSus.Tupos <> sArParev8esn then
      raise ESuvtaktikoLa8os.CreateFmt(sDevAvoijeParev8esn, [prajn.Kwdikas], dk.TrSus.Perioxn);
    dk.Epomevo;                                                  //   
    orismata.Eisagwgn(Ekfrasn(dk));           //     sse   
    fTeliko := dk.TrSus;
    if fTeliko.Tupos <> sDeParev8esn then
      raise ESuvtaktikoLa8os.CreateFmt(sDevEkleiseParev8esn, [prajn.Kwdikas], dk.TrSus.Perioxn);
    dk.Epomevo;
  end
  else begin
    orismata.Eisagwgn(sse(dk));
    fTeliko := orismata[0].Teliko;
  end;
  fTuposDedomevwv := varEmpty;
  case prajn.Tupos of
    sOxi:
      if orismata[0].TuposDedomevwv = varBoolean then
        fTuposDedomevwv := varBoolean;
    sA_M:
      if orismata[0].TuposDedomevwv in [varInt64, varDouble] then
        fTuposDedomevwv := varInt64;
    sA_T, sMovadiaioSuv, sDuadikoSuv, sMovadiaioMeiov, sDuadikoPlnv:
      if orismata[0].TuposDedomevwv in [varInt64, varDouble] then
        fTuposDedomevwv := orismata[0].TuposDedomevwv;
    sE, sEF, sHM, sLOG, sSYNnm, sT_R:
      if orismata[0].TuposDedomevwv in [varInt64, varDouble] then
        fTuposDedomevwv := varDouble;
  end;
  if fTuposDedomevwv = varEmpty then
    if suvartnsn then
      raise ESuvtaktikoLa8os.CreateFmt(sAsumbatoOrismaSuvartnsns,
        [prajn.Kwdikas, orismata[0].KwdikasKaiPerigrafn], orismata[0].Perioxn)
    else
      raise ESuvtaktikoLa8os.CreateFmt(sAsumbatoOrismaTelestn,
        [prajn.Kwdikas, orismata[0].KwdikasKaiPerigrafn], orismata[0].Perioxn);
end;

function TMovadiaiaPrajn.TuposDedomevwv: TVarType;
begin
  result := fTuposDedomevwv;
end;

{ TDuadiknPrajn }

function TDuadiknPrajn.DesTimn: Variant;
{***************************************************************************************************
              
 .      .
      result  .
***************************************************************************************************}
function EukleidioYpoloipo(a, b: Variant): Variant;
begin
  Result := a mod b;
  if Result < 0 then
    Inc(Result, Abs(b));
end;

resourcestring
  sSfalmaYpologismouDuadiknsPrajns =
    ' %s      %s.  :' + WBR
    + '%s = %s' + WBR + '%s = %s';
var
  timn1, timn2: Variant;
  meAvoxn: boolean;
  avoxn: double;
begin
  meAvoxn := (orismata[0].TuposDedomevwv = varDouble) or (orismata[1].TuposDedomevwv = varDouble);
  if meAvoxn then
    avoxn := 0.5*Power(0.1, ru8miseis[ruSnmavtikaDekadika].timn)
  else
    avoxn := 0;             //     might have not been initialized  compiler 
  timn1 := orismata[0].Timn;                    //        
// partial boolean evaluation:
  if ru8miseis[ruPlnrnsApotimnsnSuv8nkwv].timn or not (prajn.Tupos in [sH, sKai]) then
    timn2 := orismata[1].Timn;
  try
    case prajn.Tupos of
      sMovadiaioSuv, sDuadikoSuv:
        result := timn1 + timn2;
      sMovadiaioMeiov, sDuadikoPlnv:
        result := timn1 - timn2;
      sEpi:
        result := timn1 * timn2;
      sDia:
        result := timn1 / timn2;
      sMikrotero:
        if meAvoxn then
          result := timn1 - timn2 < -avoxn
        else
          result := timn1 < timn2;
      sIsov:
        if meAvoxn then
          result := abs(timn1 - timn2) <= avoxn
        else
          result := timn1 = timn2;
      sMegalutero:
        if meAvoxn then
          result := timn1 - timn2 > avoxn
        else
          result := timn1 > timn2;
      sDuvamn:
        if (fTuposDedomevwv = varInt64) then
          result := IntPower(timn1, timn2)
        else
          result := Power(double(timn1),double(timn2));
      sDiaforoANSI, sDiaforoUNICODE:
        if meAvoxn then
          result := abs(timn1 - timn2) > avoxn
        else
          result := timn1 <> timn2;
      sMikroteroHIsoANSI, sMikroteroHIsoUNICODE:
        if meAvoxn then
          result := timn1 - timn2 <= avoxn
        else
          result := timn1 <= timn2;
      sMegaluteroHIsoANSI, sMegaluteroHIsoUNICODE:
        if meAvoxn then
          result := timn1 - timn2 >= -avoxn
        else
          result := timn1 >= timn2;
      sDiv:
        if ru8miseis[ruMa8nmatikaDivMod].timn then
          result := (timn1 - EukleidioYpoloipo(timn1, timn2)) div timn2
        else
          result := timn1 div timn2;
      sMod:
        if ru8miseis[ruMa8nmatikaDivMod].timn then
          result := EukleidioYpoloipo(timn1, timn2)
        else
        result := timn1 mod timn2;
      sH:
        if (ru8miseis[ruPlnrnsApotimnsnSuv8nkwv].timn) then
          result := timn1 or timn2   //      compiler  Delphi
        else
          if timn1 then result := true
          else result := orismata[1].Timn;
      sKai:
        if (ru8miseis[ruPlnrnsApotimnsnSuv8nkwv].timn) then
          result := timn1 and timn2  //      compiler  Delphi
        else
          if not timn1 then result := false
          else result := orismata[1].Timn;
    end;
    result := VarAsType(result, fTuposDedomevwv);  //   int64   integer
  except
    on E: exception do
      raise ELa8osEktelesns.CreateFmt(sSfalmaYpologismouDuadiknsPrajns,
        [E.Message, prajn.Kwdikas, orismata[0].Kwdikas, VariantWsKeimevo(timn1),
         orismata[1].Kwdikas, VariantWsKeimevo(timn2)], prajn.Perioxn);
  end;
end;

constructor TDuadiknPrajn.Create(dk: TDiaxeiristnsKwdika; orisma1: TSustatiko;
 sse: TSuvSarEkfrasns);
{***************************************************************************************************
   +, -, *, /, div, mod, ^, =, <>, <, >, <=, >=, , .
   TrSus =  .
***************************************************************************************************}
resourcestring
  sAduvatnPrajn = '     %s    .' + WBR + '%s' + WBR + '%s';
  sOxiSuvevwsnAlfari8mntikwv = '     (%s + %s).' + WBR 
    + '    ,         .';
begin
  inherited Create;                                              //    
  fArxiko := orisma1.Arxiko;
  orismata.Eisagwgn(orisma1);
  prajn := dk.TrSus;
  dk.Epomevo;                                                                 //  
  orismata.Eisagwgn(sse(dk));                                             //   
  fTeliko := orismata[1].Teliko;
//todo:             Unicode strings
  if (orismata[0].TuposDedomevwv = varInt64) and (orismata[1].TuposDedomevwv = varInt64)
   and (prajn.Tupos in sSuv + sMeiov + [sEpi, sDuvamn, sDiv, sMod]) then
    if (prajn.Tupos <> sDuvamn)
     or orismata[1].Ypostnrizei(lsSta8ernTimn) and (orismata[1].Timn >= 0) then
      fTuposDedomevwv := varInt64
    else fTuposDedomevwv := varDouble              //2^-2   => 2^  
  else if (orismata[0].TuposDedomevwv in [varInt64, varDouble])
   and (orismata[1].TuposDedomevwv in [varInt64, varDouble])
   and (prajn.Tupos in sSuv + sMeiov + [sEpi, sDia, sDuvamn])
    then fTuposDedomevwv := varDouble
  else if (orismata[0].TuposDedomevwv = varOleStr) and (orismata[1].TuposDedomevwv = varOleStr)
   and (prajn.Tupos in sSuv) then
     if not ru8miseis[ruOxiSuvevwsnAlfari8mntikwv].timn then
       fTuposDedomevwv := varOleStr
     else
       raise ESuvtaktikoLa8os.CreateFmt(sOxiSuvevwsnAlfari8mntikwv, [orismata[0].Kwdikas,
      orismata[1].Kwdikas], prajn.Perioxn)
  else if ( (orismata[0].TuposDedomevwv = varBoolean) and (orismata[1].TuposDedomevwv = varBoolean)
     and (prajn.Tupos in sDiaforo + [sIsov, sH, sKai]) )
    or ( (orismata[0].TuposDedomevwv in [varInt64, varDouble]) and (orismata[1].TuposDedomevwv in [varInt64, varDouble])
     and (prajn.Tupos in [sIsov,sMikrotero,sMegalutero] + sDiaforo+sMikroteroHIso+sMegaluteroHIso) )
    or ( (orismata[0].TuposDedomevwv = varOleStr) and (orismata[1].TuposDedomevwv = varOleStr) and
     (prajn.Tupos in [sIsov,sMikrotero,sMegalutero] + sDiaforo+sMikroteroHIso+sMegaluteroHIso) ) 
    then fTuposDedomevwv := varBoolean
  else
    raise ESuvtaktikoLa8os.CreateFmt(sAduvatnPrajn, [prajn.Kwdikas, orismata[0].KwdikasKaiPerigrafn,
      orismata[1].KwdikasKaiPerigrafn], prajn.Perioxn);
end;

function TDuadiknPrajn.TuposDedomevwv: TVarType;
begin
  result := fTuposDedomevwv;
end;

function VariantWsKeimevo(const v: Variant): WideString;
{***************************************************************************************************
    ,    ,      
   .
     -     ,    
              .
***************************************************************************************************}
resourcestring
  sPollaStoixeiaPivaka = '            ';
var
  i, j, pln8osStoixeiwv, pln8osAvaDiastasn: integer;
  deiktes: array of integer;
begin
  if VarIsEmpty(v) then result := KevoStr
  else if (VarType(v) = varBoolean) then
    if v then result := mPKTDL(sAln8ns)
    else result := mPKTDL(sCeudns)
  else if VarType(v) = varDouble then                                        //   
    result := FloatToStrF(v, ffFixed, 15, ru8miseis[ruDekadikaCnfia].timn)
  else if VarIsArray(v) then
  begin                      // .. (((0, 1), (2, 3)), ((4, 5), (6, 7)))   2x2x2
    result := '(';
    SetLength(deiktes, VarArrayDimCount(v));
    pln8osStoixeiwv := 1;                             //      
    for i := 1 to VarArrayDimCount(v) do
      pln8osStoixeiwv := pln8osStoixeiwv * VarArrayHighBound(v, i);
    if pln8osStoixeiwv > 10000 then
      result := result + sPollaStoixeiaPivaka
    else for i := 0 to pln8osStoixeiwv - 1 do                       //     
    begin         //todo:   ,     TWideStringList     
      pln8osAvaDiastasn := i;//edit:   ( )    
      for j := VarArrayDimCount(v) downto 1 do                             //   
      begin
        deiktes[j - 1] := pln8osAvaDiastasn mod VarArrayHighBound(v, j) + 1;
        pln8osAvaDiastasn := pln8osAvaDiastasn div VarArrayHighBound(v, j);
      end;
      if not VarIsEmpty(VarArrayGet(v, deiktes)) then
        Insert(VariantWsKeimevo(VarArrayGet(v, deiktes)) + ', ', result, Length(result) + 1)
      else Insert(KevoStr + ', ', result, Length(result) + 1);
    end;
    SetLength(result, Length(result) - 1);
    result[Length(result)] := ')';
  end
  else                         // ,      
    result := VarAsType(v, varOleStr);
end;

{***************************************************************************************************
 edit:   ,       .
           la,   
  :
         ::=  |  |  | (  )
           ::=  ^  | 
    ::= -  | +  | 
       ::=   ( [*,/,div,mod] )*
             ::=  ([+,-] )*
///        ::=  ([&,^,|] )*                                    // 
            ::=  ([=,<,>,<>,<=,>=] [ | ])*
       ::=   | 
       ::=  ( )*
          ::=  ( )*

        :
   8. , , , 
   7. ^                    (    !!!  basic)
   6.  -, +
   5. *, /, div, mod
   4. +, -,
   3. < , <=, =, <>, >, >=
   2. 
   1. 
   0. 

///  -   , .. 1/-2/3:   -   
///  /       1/-(2/3); ,     /   
///            1/(-2)/3.
///         -      
+-----------------------------------------------------+
|             Delphi                     |
+-----------------------------------+-----------------+
| @, not                            | first (highest) |
| *, /, div, mod, and, shl, shr, as | second          |
| +, , or, xor                     | third           |
| =, <>, <, >, <=, >=, in, is       | fourth (lowest) |
+-----------------------------------+-----------------+

+------------------------------------+-------------------------------+---------+
|                             Basic                               |
+------------------------------------+-------------------------------+---------+
| Arithmetic                         | Comparison                    | Logical |
+------------------------------------+-------------------------------+---------+
| Exponentiation (^)                 | Equality (=)                  | Not     |
| Negation (-)                       | Inequality (<>)               | And     |
| Multiplication and division (*, /) | Less than (<)                 | Or      |
| Integer division (\)               | Greater than (>)              | Xor     |
| Modulus arithmetic (Mod)           | Less than or equal to (<=)    | Eqv     |
| Addition and subtraction (+, -)    | Greater than or equal to (>=) | Imp     |
| String concatenation (&)           | Like                          |         |
|                                    | Is                            |         |
+------------------------------------+-------------------------------+---------+

      :
+----------------+------+---+--------------+-----------------------------------+
|         |  |  |  |                          |
+----------------+------+---+--------------+-----------------------------------+
|               | 1    |  |     |                          |
| [2*]         | 2.23 |  |      |                            |
| (2, 3) | -    |  |        |                            |
|  !   | -    | - |        |   ""  ... |
+----------------+------+---+--------------+-----------------------------------+
 ""   drop-down,      ( )   edit.
       edit,          .
    toggle          .
 ""    .    :
 * , ,    (TEkfrasnAvagvwristikou)
 * 
 * 
 * " "

+----------------+
|   |
+----------------+
  Ekfrasn     dk.Epomevo!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      dk.Epomevo    ,      
     . :
 *    TrSus     ,      
    Stoixeio  dk.Epomevo
 *    TrSus     ,       
   .
 *     result,      ""   object.  
     
     result := TDuadiknPrajn.Create(result, Prosnmasmevos)
        Prosnmasmevos    Create,  
     result := TDuadiknPrajn.Create(result, @Prosnmasmevos)
         result      .
+-----------+
|  |
+-----------+
 , .      :
 *  Stoixeio  Ekfrasn,    dk.Epomevo,   ,   
   result.
 *  DuadiknPrajn    ,      orismata 
      (.. result := TDuadiknPrajn.Create(result, @Prosnmasmevos).
        dk.TrSus =     .
***************************************************************************************************}

function Stoixeio(dk: TDiaxeiristnsKwdika): TSustatiko;
{***************************************************************************************************
         ::=  |  |  | (  )
***************************************************************************************************}
resourcestring
  sLa8osStoixeio = '  %s    ';
  sAprosmevoTelosEkfrasns = '   ,       ';
begin
  if dk.TrSus.Tupos = sArParev8esn then
    result := TEkfrasnSeParev8esn.Create(dk)
  else if dk.TrSus.Tupos in [sA_M, sA_T, sE, sEF, sHM, sLOG, sSYNnm, sT_R] then
    result := TMovadiaiaPrajn.Create(dk, @Ekfrasn)
  else if dk.TrSus.Tupos in [sAkeraiaSt, sPragmatiknSt, sAlfari8mntiknSt, sAln8ns, sCeudns] then
  begin
    result := dk.TrSus;
    (result as TSta8era).Apotimnse;                 //      
    dk.Epomevo;
//           , 
//1.     ,     objects
//2.    ,   exception      dk.
//..  <- (1 + 2    & )
// *      ,
// *       1,
// *          1,
// *  exception  2 ->Epomevo,
// *    destructors     . !
  end
  else if dk.TrSus is TEkfrasnAvagvwristikou then
    if (dk.TrSus as TEkfrasnAvagvwristikou).avagvwristiko <> nil then
      result := (dk.TrSus as TEkfrasnAvagvwristikou).avagvwristiko.SarwseEkfrasn(dk)
    else raise ESuvtaktikoLa8os.CreateFmt(sAgvwstoAvagvwristiko, [dk.TrSus.Kwdikas], dk.TrSus.Perioxn)
  else if (dk.TrSus is TTelosGrammns) then                   // sTelosGrammns  sTelosArxeiou
    raise ESuvtaktikoLa8os.Create(sAprosmevoTelosEkfrasns, dk.TrSus.Perioxn)
  else
    raise ESuvtaktikoLa8os.CreateFmt(sLa8osStoixeio, [dk.TrSus.Kwdikas], dk.TrSus.Perioxn);
end;

function ProsnmasmevosXwrisDuvamn(dk: TDiaxeiristnsKwdika): TSustatiko;
{***************************************************************************************************
    ::= -  | +  | 
***************************************************************************************************}
begin
  if dk.TrSus.Tupos in sSuv + sMeiov then
    result := TMovadiaiaPrajn.Create(dk, @ProsnmasmevosXwrisDuvamn)
  else result := Stoixeio(dk);
end;

function Prosnmasmevos(dk: TDiaxeiristnsKwdika): TSustatiko; forward;

function Duvamn(dk: TDiaxeiristnsKwdika): TSustatiko;
{***************************************************************************************************
           ::=  ^  | 
***************************************************************************************************}
var
  sse: TSuvSarEkfrasns;
begin
  if ru8miseis[ruAristernProsetairistikotntaDuvamns].timn then
    sse := @ProsnmasmevosXwrisDuvamn
  else                   
    sse := @Prosnmasmevos;
  result := Stoixeio(dk);
  while dk.TrSus.Tupos = sDuvamn do             
    result := TDuadiknPrajn.Create(dk, result, sse);
end;

function Prosnmasmevos(dk: TDiaxeiristnsKwdika): TSustatiko;
{***************************************************************************************************
    ::= -  | +  | 
***************************************************************************************************}
begin
  if dk.TrSus.Tupos in sSuv + sMeiov then
    result := TMovadiaiaPrajn.Create(dk, @Prosnmasmevos)
  else result := Duvamn(dk);
end;

function Paragovtas(dk: TDiaxeiristnsKwdika): TSustatiko;
{***************************************************************************************************
       ::=   ( [*,/,div,mod] )*
***************************************************************************************************}
begin
  result := Prosnmasmevos(dk);
  while dk.TrSus.Tupos in [sEpi, sDia, sDiv, sMod] do
    result := TDuadiknPrajn.Create(dk, result, @Prosnmasmevos);
end;

function Oros(dk: TDiaxeiristnsKwdika): TSustatiko;
{***************************************************************************************************
             ::=  ([+,-] )*
***************************************************************************************************}
begin
  result := Paragovtas(dk);
  while dk.TrSus.Tupos in sSuv + sMeiov do
    result := TDuadiknPrajn.Create(dk, result, @Paragovtas);
end;

function EkfrasnOxi(dk: TDiaxeiristnsKwdika): TSustatiko; forward;

function Sxesn(dk: TDiaxeiristnsKwdika): TSustatiko;
{***************************************************************************************************
            ::=  ([=,<,>,<>,<=,>=] [ | ])*
***************************************************************************************************}
begin
  result := Oros(dk);
  while dk.TrSus.Tupos in
   [sIsov, sMikrotero, sMegalutero] + sDiaforo + sMikroteroHIso + sMegaluteroHIso do
    if dk.EpSus.Tupos = sOxi then
     result := TDuadiknPrajn.Create(dk, result, @EkfrasnOxi)
    else result := TDuadiknPrajn.Create(dk, result, @Oros);
end;

function EkfrasnOxi(dk: TDiaxeiristnsKwdika): TSustatiko;
{***************************************************************************************************
       ::=   | 
***************************************************************************************************}
begin
  if (dk.TrSus.Tupos = sOxi) then
    result := TMovadiaiaPrajn.Create(dk, @EkfrasnOxi)
  else result := Sxesn(dk);
end;

function EkfrasnKai(dk: TDiaxeiristnsKwdika): TSustatiko;
{***************************************************************************************************
       ::=  ( )*
***************************************************************************************************}
begin
  result := EkfrasnOxi(dk);
  while (dk.TrSus.Tupos = sKai) do
    result := TDuadiknPrajn.Create(dk, result, @EkfrasnOxi);
end;

function Ekfrasn(dk: TDiaxeiristnsKwdika): TSustatiko;
{***************************************************************************************************
          ::=  ( )*
***************************************************************************************************}
begin
  if ru8miseis[ruKaiProngeitaiH].timn then
  begin
    result := EkfrasnKai(dk);
    while (dk.TrSus.Tupos = sH) do
      result := TDuadiknPrajn.Create(dk, result, @EkfrasnKai);
  end
  else begin
    result := EkfrasnOxi(dk);
    while (dk.TrSus.Tupos in [sH, sKai]) do
      result := TDuadiknPrajn.Create(dk, result, @EkfrasnOxi);
  end;
end;

end.
