{***************************************************************************************************
 This file is part of the GLOSSA Programming Language Interpreter (  )
 Copyright (C) 2002-2008 Alkis Georgopoulos - http://users.sch.gr/alkisg/

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************************************}
unit uEvtoles;
{***************************************************************************************************
     TAfnrnmevoProgramma        
     .      ,  .  
   TSarwtnsEkfrasewv      TEkfrasn       
 (    )         
   .           
 . ..
     > 4      //       TAfnrnmevoProgramma   << 
   _  << 32        //      _     .
***************************************************************************************************}
interface

uses
  Variants, Math, SysUtils, Windows, uLektikosAvalutns, uSustatika, uEkfraseis, uRu8miseis, uLa8os;

type
  TStoiba = class
  private
    fSP: integer;
    fStoixeia: array of Variant;
    function    DesStoixeio(p: integer): PVariant;
    procedure   QeseStoixeio(p: integer; const v: PVariant);
  public
    BP: integer;                                           //   Stoixeia   BP
    property    SP: integer read fSP;             // SP    ,   = 0
    property    Stoixeia[p: integer]: PVariant read DesStoixeio write QeseStoixeio; default;
    constructor Create;                                                      //  fStoixeia
    procedure   Arxikopoinse;                      //  ,   
    procedure   Push(const v: Variant);
    function    Pop: Variant;       //;  variant     Length(fStoixeia)
    destructor  Destroy; override;
  end;

  TAfnrnmevoProgramma = class;

  TEvtoln = class(TSuv8etoSustatiko)
  protected
    p: TAfnrnmevoProgramma;
    orismata: TListaSustatikwv;                          //      
    nEpomevnEvtoln: TEvtoln;                //        
    procedure   Avagvwrise; virtual; abstract;                        //   Create
  public
    constructor Create(p: TAfnrnmevoProgramma);    //    override   
    function    Ektelesn: TEvtoln; virtual;   // p.ArxnEvtolns    epomevnEvtoln
    procedure   ProsperaseTelosGrammns;                          //    
    destructor  Destroy; override;                                         //  orismata
  end;

  TEvtolnAmesnsEktelesns = class(TEvtoln);       // p.ArxnEvtolns    ' 

  TEvtolnDiakladwsns = class(TEvtoln)                        //    ,  
{***************************************************************************************************
  TEvtolnDiakladwsns       ,    
 override       inherited.
      orismata[0],     override  Prospa8nseApotimnsn
 (   KaveApotimnsn)   orismata[0]       , . 
    Ektelesn  (  try block)  Prospa8nseApotimnsn,     
 (    )       fApotimnsn, 
 fApotimnsn := unassigned.
    ArxnEvtolns,     fApotimnsn = unassigned  
 KaveApotimnsn,         .
    ArxnEvtolns.
***************************************************************************************************}
  protected
    fApotimnsn: Variant;                                                 //unassigned  true  false
    nEvtolnDiakladwsns: TEvtoln;                           // fApotimnsn = false  
    procedure   Prospa8nseApotimnsn; virtual;
    procedure   KaveApotimnsn; virtual;
  public                                                                   //orismata[0] =  
    property    Apotimnsn: Variant read fApotimnsn;              //    
//todo:    write permissions,          
    function    Ektelesn: TEvtoln; override;                         //   
  end;

  TEvtolnAva8esnTimns = class(TEvtoln)
  protected        //orismata[0] =     , orismata[1] =  
    procedure   Avagvwrise; override;
  public
    function    Ektelesn: TEvtoln; override;
  end;

  TEvtolnDiabase = class(TEvtoln)
  protected                                                          //orismata[0..] =  
    procedure   Avagvwrise; override;
  public
    function    Ektelesn: TEvtoln; override;
  end;

  TEvtolnGraye = class(TEvtoln)
  protected                                                          //orismata[0..] =  
    procedure   Avagvwrise; override;
  public
    function    Ektelesn: TEvtoln; override;
  end;

  TEvtolnAv = class(TEvtolnDiakladwsns)
  protected                                                                //orismata[0] =  
    nTeleutaiaAlliwsAv: TEvtolnDiakladwsns;//       _
    evtolesGiaTelosAv: array of TEvtoln;//        
    procedure   Avagvwrise; override;
    procedure   SuvdeseAlliws(nApopavwEvtoln, evtolnAlliws: TEvtoln);
    procedure   SuvdeseTelosAv(evtolnTelosAv: TEvtoln);
  end;                         // Ektelesn   ,       Avagvwrise

  TEvtolnAlliwsAv = class(TEvtolnDiakladwsns)
  protected
    procedure   Avagvwrise; override;
  end;

  TEvtolnAlliws = class(TEvtoln)
  protected  
    procedure   Avagvwrise; override;
  end;

  TEvtolnTelosAv = class(TEvtoln)
  protected
    procedure   Avagvwrise; override;
  end;

  TEvtolnEpileje = class(TEvtoln)
  private                                                      //orismata[0] =   - 
    nTeleutaiaPeriptwsn: TEvtoln;                                 //    
    evtolesGiaTelosEpilogwv: array of TEvtoln;                        //  _
  protected
    apotimnsn: Variant;                //   ,    ""
    procedure   Avagvwrise; override;
    procedure   SuvdesePeriptwsn(nApopavwEvtoln: TEvtoln; evtolnPeriptwsn: TEvtoln);
    procedure   SuvdeseTelosEpilogwv(evtolnTelosEpilogwv: TEvtoln);
  public
    function    Ektelesn: TEvtoln; override;                      //    
  end;

  TEvtolnPeriptwsn = class(TEvtolnDiakladwsns)
{***************************************************************************************************
  TEvtolnPeriptwsn  override  inherited Prospa8nseApotimnsn  TEvtolnDiakladwsns, 
   ,        .
  orismata[0..]   ,      prajeis  
   .      .
                (
    ).
***************************************************************************************************}
  private
    nEvtolnEpileje: TEvtolnEpileje;
    prajeis: array of TTuposSustatikou;  //  ,     sDiastnma
    procedure   ElegjeEkfrasn(ekf: TSustatiko);                     //   
    procedure   SarwseEkfraseis;
  protected                                            //orismata[0..] =    
    procedure   Prospa8nseApotimnsn; override;
    procedure   Avagvwrise; override;
  end;

  TEvtolnPeriptwsnAlliws = class(TEvtoln)//       
  protected
    procedure   Avagvwrise; override;
  end;

  TEvtolnTelosEpilogwv = class(TEvtoln)
  protected
    procedure   Avagvwrise; override;
  end;

  TEvtolnOso = class(TEvtolnDiakladwsns)
  protected
    procedure   Avagvwrise; override;
  end;

  TEvtolnGia = class;                                      //   TEvtolnMexriMeBnma

  TEvtolnMexriMeBnma = class(TEvtolnDiakladwsns)
{***************************************************************************************************
        . '    ,   
    (= TEvtolnGiaApo)           
  (= TEvtolnMexriMeBnma).      ,    
   .     override   Prospa8nseApotimnsn   KaveApotimnsn.
  Prospa8nseApotimnsn    try  KaveApotimnsn.
  KaveApotimnsn         ,   
       ,       .
       ,  TEvtolnGia  varEmpty  timnMexri.
***************************************************************************************************}
  private
    timnMexri, timnBnmatos: Variant;                         //    
    prajn: TTuposSustatikou;                            //>=  <=,      
    nEvtolnGia: TEvtolnGia;                                             //    
  protected                 //orismata[0] =  , orismata[1] =  _,  
    procedure   Prospa8nseApotimnsn; override;
    procedure   KaveApotimnsn; override;
    procedure   Avagvwrise; override;
  end;

  TEvtolnGia = class(TEvtolnAmesnsEktelesns)
  private
    nEvtolnMexriMeBnma: TEvtolnMexriMeBnma;
  protected
    procedure   Avagvwrise; override;
  public
    function   Ektelesn: TEvtoln; override;
  end;

  TEvtolnTelosEpavalnyns = class(TEvtoln)                           //     
  protected
    procedure   Avagvwrise; override;
  end;

  TEvtolnArxnEpavalnyns = class(TEvtoln)
  protected
    procedure   Avagvwrise; override;
  end;

  TEvtolnMexrisOtou = class(TEvtolnDiakladwsns)
  protected
    procedure   Avagvwrise; override;
  end;

  TEvtolnKalese = class(TEvtoln)
  protected
    procedure   Avagvwrise; override;                     //,  TEkfrasnDiadikasias.Sarwse
  public
    function    Ektelesn: TEvtoln; override;
  end;

  TEvtolnProgramma = class(TEvtolnAmesnsEktelesns)
  protected
    procedure   Avagvwrise; override;
  end;

  TEvtolnTelosProgrammatos = class(TEvtoln)
  protected
    procedure   Avagvwrise; override;
  end;

  TEvtolnDiadikasia = class(TEvtolnProgramma)
  protected
    procedure   Avagvwrise; override;
  end;

  TEvtolnTelosDiadikasias = class(TEvtoln)
  protected
    procedure   Avagvwrise; override;
  end;

  TEvtolnSuvartnsn = class(TEvtolnDiadikasia)
  protected
    procedure   Avagvwrise; override;
  end;

  TEvtolnTelosSuvartnsns = class(TEvtoln)
  protected
    procedure   Avagvwrise; override;
  end;

  TEvtolnDnlwseis = class(TEvtolnAmesnsEktelesns)
  protected
    procedure   Avagvwrise; override;
  end;

  TOvomasmevnSta8era = class(TAvagvwristiko)
  protected                                                     //  TrSus =  
    fTimn: Variant;
    function    Leitourgies: Word; override; //: lsDesTimn or lsSta8ernTimn or lsSta8ernApotimnsn
    function    DesTimn: Variant; override;
  public
    constructor Create(p: TAfnrnmevoProgramma);               //     
    function    Perigrafn: WideString; override;                         //..  
    function    TuposDedomevwv: TVarType; override;
// class function SarwseEkfrasn    TAvagvwristiko
  end;

//TEkfrasnOvomasmevnsSta8eras  ,    TEkfrasnAvagvwristikou

  TMetablntn = class(TAvagvwristiko)
  protected                                                   //  TrSus =  
    fMetatopisn: integer;
    stoiba: TStoiba;
    fTuposDedomevwv: TVarType;
    function    Leitourgies: Word; override;    //: lsDesTimn or lsQeseTimn or lsSta8ernApotimnsn
    function    DesTimn: Variant; override;
    procedure   QeseTimn(const t: Variant); override;
  public
    property    Metatopisn: integer read fMetatopisn;
    constructor Create(p: TAfnrnmevoProgramma; tupos: TVarType; metatopisn: integer);
    function    Perigrafn: WideString; override;                          //..  ޻
    function    TuposDedomevwv: TVarType; override;     //varInt64, varDouble, varOleStr, varBoolean
    procedure   Arxikopoinse; virtual;  //  (  )    varEmpty
  end;

  TMetablntnPivakas = class(TMetablntn)
  protected                                          //  TrSus =   - 
    procedure   QeseTimn(const t: Variant); override;         //     varArray
  public
    oria: array of integer;                                           //   (=1)   
    constructor Create(p: TAfnrnmevoProgramma; tupos: TVarType; metatopisn: integer);
    function    Perigrafn: WideString; override;                           //..  
    procedure   Arxikopoinse; override; //  (  )    varEmpty
    class function SarwseEkfrasn(dk: TDiaxeiristnsKwdika): TSustatiko; override;
  end;

//TEkfrasnMetablntnsPivaka  ,    TEkfrasnAvagvwristikou

  TEkfrasnStoixeiouPivaka = class(TSuv8etnEkfrasn)                                  //.. [2*, 1]
//    TEkfrasnAvagvwristikou,        Create
  protected                                         //orismata =      
    function    DesTimnHUnassigned: Variant;  //      exception
    function    DesTimn: Variant; override;
    procedure   QeseTimn(const t: Variant); override;         //     
    function    Leitourgies: Word; override;    //: lsDesTimn or lsQeseTimn or lsSta8ernApotimnsn
  public
    metablntnPivakas: TMetablntnPivakas;
    constructor Create(dk: TDiaxeiristnsKwdika);
    function    Perigrafn: WideString; override;                     //..   
    function    TuposDedomevwv: TVarType; override;        //     varArray
  end;

  TYpoprogramma = class(TAvagvwristiko)
{***************************************************************************************************
    2 pass compiler,     :
        dk   .   ,
         .        
   ,       breakpoints.
        TProgramma,        
   ,  ,   .   
 TProgramma.trexwvYpoprogramma.
***************************************************************************************************}
  private
    parametroi: array of TEkfrasnAvagvwristikou;        //      
    fAvagvwristika: TListaSustatikwv;                // ,   .Create  
    evtoles: array of TEvtoln;                                     //     
    stoibaEmfwleumevwv: array of TEvtoln;                    //  , , , _
  protected
    p: TAfnrnmevoProgramma;
    te: TEvtoln;       //TrexousaEvtoln,         
    procedure   EisagwgnAvagvwristikou(av: TAvagvwristiko); //    
    function    AvazntnsnAvagvwristikou(ovoma: WideString): TAvagvwristiko;
    procedure   Eisagwgn(evtoln: TEvtoln);                          //     
    procedure   W8nsn(evtoln: TEvtoln);                                //   
    function    Ejw8nsn: TEvtoln;           //       
    function    Krufokoitagma: TEvtoln;//        
    procedure   EgkuroOvoma(ea: TEkfrasnAvagvwristikou);
    procedure   SarwseEvtolnYpoprogramma; virtual; abstract;            //..  (, )
    procedure   AvaluseParametrous; virtual;                   //   
  public
    property    Avagvwristika: TListaSustatikwv read fAvagvwristika;
    constructor Create(p: TAfnrnmevoProgramma);                //     
    destructor  Destroy; override;
    procedure   SarwseSwma;                                         //    _
    procedure   Klnsn; // ,       .Ektelesn
    class function SarwseEkfrasn(dk: TDiaxeiristnsKwdika): TSustatiko; override;    // 
  end;

  TEkfrasnYpoprogrammatos = class(TSuv8etnEkfrasn)                        //   
  public
    upoprogramma: TYpoprogramma;
  end;

  TKuriwsProgramma = class(TYpoprogramma)
  protected
    procedure   SarwseEvtolnYpoprogramma; override;                            //..  
  public
    function    Perigrafn: WideString; override;                                       //= 
  end;

  TEkfrasnKuriwsProgrammatos = class(TEkfrasnYpoprogrammatos)
  public
    constructor Create(dk: TDiaxeiristnsKwdika);
  end;                                      //       !!!

  TDiadikasia = class(TYpoprogramma)
//          (: , , , ).
//    ,         
//.
  protected
    procedure   SarwseEvtolnYpoprogramma; override;
    procedure   AvaluseParametrous; override;        //      
  public
    function    Perigrafn: WideString; override;                                      //= 
  end;

  TEkfrasnDiadikasias = class(TEkfrasnYpoprogrammatos)               //.. (, )
//    TEkfrasnAvagvwristikou,        Create
  public                                                    //orismata =    
    constructor Create(dk: TDiaxeiristnsKwdika);     //    override  Perigrafn
  end;      //  Leitougries,           

  TSuvartnsn = class(TDiadikasia)
//     ,    lsDesTimn   lsQeseTimn.
//          (=result),     
// .            .
  protected                        //     = -   - 1
    fTuposDedomevwv: TVarType;                   //   TSuvartnsn.TuposDedomevwv
    function    Leitourgies: Word; override;    //: lsDesTimn or lsQeseTimn or lsSta8ernApotimnsn
    function    DesTimn: Variant; override;       //     
    procedure   QeseTimn(const t: Variant); override;                 //  stoiba[metatopisn]^
    procedure   SarwseEvtolnYpoprogramma; override;
  public
    function    Perigrafn: WideString; override;
    function    TuposDedomevwv: TVarType; override;
    class function SarwseEkfrasn(dk: TDiaxeiristnsKwdika): TSustatiko; override;
  end;

  TEkfrasnSuvartnsns = class(TEkfrasnDiadikasias)
{***************************************************************************************************
 ..  2 * (3, a[4, 5])
 -  Timn     
   * stoiba.Push(empty)
   * Suvartnsn.Klnsn
   * stoiba.Pop(result)
***************************************************************************************************}
  protected
    function    DesTimn: Variant; override;                                    //  
    function    Leitourgies: Word; override;                                   //:  lsDesTimn
  public
//constructor Create(dk: TDiaxeiristnsKwdika)  ,   
    function    TuposDedomevwv: TVarType; override;
  end;

  TKln8evYpoprogramma = record
    upoprogramma: TYpoprogramma;
    BP: integer;
  end;

  TKln8evtaYpoprogrammata = array of TKln8evYpoprogramma;

  TAfnrnmevoProgramma = class
{***************************************************************************************************
  TAfnrnmevoProgramma       TKuriwsProgramma.   , 
   .
      stack   ,       uMain 
         .
***************************************************************************************************}
  protected
    dk: TDiaxeiristnsKwdika;
    upoprogrammata: array of TYpoprogramma;
    fStoiba: TStoiba;
    callstack: TKln8evtaYpoprogrammata;
    fKP: TKuriwsProgramma;             // KuriwsProgramma       
    fTY: TYpoprogramma;                                                         //TrexwvYpoprogramma
    seAva8esnTimns: boolean;         //       
    procedure   Ejodos(const keimevo: WideString); virtual; abstract;
    procedure   Eisodos(ekfMet: TSustatiko); virtual; abstract;
    procedure   ArxnEvtolns; virtual; abstract;
    function    DesTE: TEvtoln;
    procedure   Mndevise;                       //       Destroy
  public                                            //    , ,  .
    ovoma: WideString;                               //      
    property    TE: TEvtoln read DesTE;
    property    TY: TYpoprogramma read fTY;
    property    KP: TKuriwsProgramma read fKP;
    property    Stoiba: TStoiba read fStoiba;
    constructor Create;
    destructor  Destroy; override;
    procedure   Metaglwttisn;
    procedure   Ektelesn; virtual;
    procedure   DiadikasiaAvalusnsAvagvwristikou(eav: TEkfrasnAvagvwristikou);
  end;

{
      :
-----------------------------------------
 1.      metatopisn       
     bp   .  bp        , 
          bp.     bp     
             .
                
               
       bp    .
 2.           ,  
      (bp / ).    ,     TEvtolnProgramma,
    TEvtolnDiadikasia  TEvtolnSuvartnsn.      
     .           
            .
}

{**************************************************************************************************}
implementation
{**************************************************************************************************}

{ TStoiba }

function TStoiba.DesStoixeio(p: integer): PVariant;
begin
  if (BP + p >= Low(fStoixeia)) and (BP + p <= High(fStoixeia)) then
    result := @fStoixeia[BP + p]
  else result := nil;
end;

procedure TStoiba.QeseStoixeio(p: integer; const v: PVariant);
begin
  fStoixeia[BP + p] := v^;                    //    VarClear   
end;

constructor TStoiba.Create;
begin
//  ,   .  SetLength(fStoixeia, 0);
end;

procedure TStoiba.Arxikopoinse;
var
  i: integer;
begin
  for i := Low(fStoixeia) to High(fStoixeia) do
    VarClear(fStoixeia[i]);                                       //    
  fStoixeia := nil;
  BP := 0;
  fSP := 0;
end;

procedure TStoiba.Push(const v: Variant);
begin
  if Length(fStoixeia) < SP + 1 then
    SetLength(fStoixeia, Length(fStoixeia) + 10);                  //   10 
  fStoixeia[SP] := v;
  Inc(fSP);
end;

function TStoiba.Pop: Variant;
{***************************************************************************************************
  Pop   VarClear,           uMain   
      . VarClear    Destroy.
***************************************************************************************************}
begin
  Dec(fSP);
  result := fStoixeia[SP];
end;

destructor TStoiba.Destroy;
begin
  Arxikopoinse;
  inherited;
end;

constructor TEvtoln.Create(p: TAfnrnmevoProgramma);
{***************************************************************************************************
  p.TY.Eisagwgn   nEpomevnEvtoln         '
   .
***************************************************************************************************}
begin
  inherited Create;
  self.p := p;
  fArxiko := p.dk.TrSus;
  orismata := TListaSustatikwv.Create;
  if self is TEvtolnGia then        //workaround   Gia      MexriMeBnma
  begin
    p.TY.Eisagwgn(self);
    try         //   workaround,       Avagvwrise 
      Avagvwrise;// TY      TEvtolnGia,       
      fTeliko := p.dk.TrSus;            //  ,       Enter
      ProsperaseTelosGrammns;
    except     //Create     Free     Delphi (   TEvtoln.Destroy)
//  exception    ProsperaseTelosGrammns (    
//   1  10 _),     .     ,
//    ,   .    if    
//  ,      p.TY.evtoles      .
      if p.TY.evtoles[High(p.TY.evtoles)] is TEvtolnMexriMeBnma then
        p.TY.evtoles[High(p.TY.evtoles) - 1] := p.TY.evtoles[High(p.TY.evtoles)];
      SetLength(p.TY.evtoles, Length(p.TY.evtoles) - 1);         //  ,  
      raise;
    end;
  end else
  begin
    Avagvwrise;
    fTeliko := p.dk.TrSus;              //  ,       Enter
// Avagvwrise  TEvtolnGia     TEvtolnMexriMeBnma.Create,  
//  ProsperaseTelosGrammns   TEvtolnMexriMeBnma     TEvtolnGia.
    if not (self is TEvtolnMexriMeBnma) then
      ProsperaseTelosGrammns;
    p.TY.Eisagwgn(self);           //      Avagvwrise   tu.te
  end;
end;

function TEvtoln.Ektelesn: TEvtoln;
{***************************************************************************************************
    ,      .
     ,            
  ,  ,     .
***************************************************************************************************}
begin
  p.ArxnEvtolns;                           //     , breakpoint 
  result := nEpomevnEvtoln;
end;

procedure TEvtoln.ProsperaseTelosGrammns;
{***************************************************************************************************
         Enter.      TrSus =
 Enter.    Enter  ( ),    .
        .
***************************************************************************************************}
resourcestring
  sHEvtolnTeleiwse = '  ,   .';
begin
  if not (p.dk.TrSus is TTelosGrammns) then
    raise ESuvtaktikoLa8os.Create(sHEvtolnTeleiwse, p.dk.TrSus.Perioxn);
  repeat
    fTeliko := p.dk.TrSus;
  until p.dk.Epomevo.Tupos <> sTelosGrammns;
end;

destructor TEvtoln.Destroy;
begin
  FreeAndNil(orismata);       // TListaSustatikwv.Destroy     
  inherited;
end;

procedure TEvtolnDiakladwsns.Prospa8nseApotimnsn;
begin
  if orismata[0].Ypostnrizei(lsSta8ernApotimnsn) then
  try                                                      //     
    fApotimnsn := orismata[0].Timn;
  finally                                               //  ,   
  end;
end;

procedure TEvtolnDiakladwsns.KaveApotimnsn;
{***************************************************************************************************
     fApotimnsn = UnAssigned.
***************************************************************************************************}
begin
  fApotimnsn := orismata[0].Timn;               //         
end;

function TEvtolnDiakladwsns.Ektelesn: TEvtoln;
{***************************************************************************************************
   inherited ,      = orismata[0].
***************************************************************************************************}
begin
  VarClear(fApotimnsn);
  Prospa8nseApotimnsn;                                               //    override
  p.ArxnEvtolns;                        //  fApotimnsn       
  if VarIsEmpty(fApotimnsn) then            //        
  begin
    KaveApotimnsn;
    p.ArxnEvtolns;                         //  ,    
  end;
  if fApotimnsn then
    result := nEpomevnEvtoln
  else result := nEvtolnDiakladwsns;
end;

procedure TEvtolnAva8esnTimns.Avagvwrise;
{***************************************************************************************************
      .        -
 ,               .
***************************************************************************************************}
resourcestring
  sPerimevaAva8esnTimns = '    ';
  sAduvatnAva8esnTimns = '         .' + WBR + '%s';
  sAva8esnDiaforetikwvTupwv = '         '
     + '      .' + WBR + '%s' + WBR + '%s';
begin
  p.seAva8esnTimns := true;   //  TSuvartnsn.SarwseEkfrasn   ,  
  orismata.Eisagwgn(Ekfrasn(p.dk));
  p.seAva8esnTimns := false;
  if not (p.dk.TrSus.Tupos in sAva8esnTimns) then
    raise ESuvtaktikoLa8os.Create(sPerimevaAva8esnTimns, p.dk.TrSus.Perioxn);
  if not orismata[0].Ypostnrizei(lsQeseTimn)
   or not (orismata[0].TuposDedomevwv in [varInt64, varDouble, varOleStr, varBoolean]) then
    raise ESuvtaktikoLa8os.CreateFmt(sAduvatnAva8esnTimns, [orismata[0].KwdikasKaiPerigrafn],
      orismata[0].Perioxn);
  p.dk.Epomevo;                                                                  //  <-
  orismata.Eisagwgn(Ekfrasn(p.dk));
  fTeliko := p.dk.TrSus;
  if (orismata[0].TuposDedomevwv <> orismata[1].TuposDedomevwv) and
//         varInt64  varDouble
   not ((orismata[0].TuposDedomevwv = varDouble) and (orismata[1].TuposDedomevwv = varInt64)) then
    raise ESuvtaktikoLa8os.CreateFmt(sAva8esnDiaforetikwvTupwv,
      [orismata[0].KwdikasKaiPerigrafn, orismata[1].KwdikasKaiPerigrafn], Perioxn);
end;

function TEvtolnAva8esnTimns.Ektelesn: TEvtoln;
begin
  result := inherited Ektelesn;                //       
  orismata[0].Timn := orismata[1].Timn;                    //       
end;

procedure TEvtolnDiabase.Avagvwrise;
{***************************************************************************************************
       .     
   .
   ,    Enter
***************************************************************************************************}
resourcestring
  sAduvatoDiabasmaEkfrasns = '       %s';
  sAduvatoDiabasmaLogikwv = '     .';
  sOxiDiabaseSeSuvartnseis = '       ,  '
    + '             '
    + ' .            '
    + ' .';
var
  i: integer;
begin
  if p.TY.evtoles[0] is TEvtolnSuvartnsn then
    if ru8miseis[ruOxiGraceDiabaseSeSuvartnseis].timn then
      raise ESuvtaktikoLa8os.Create(sOxiDiabaseSeSuvartnseis, p.dk.TrSus.Perioxn);
  i := 0;
  if p.dk.EpSus is TTelosGrammns then                                   //  
    p.dk.Epomevo                                                    //   
  else repeat
    p.dk.Epomevo;                                        //      
    orismata.Eisagwgn(Ekfrasn(p.dk));
    if not orismata[i].Ypostnrizei(lsQeseTimn)
     or not (orismata[i].TuposDedomevwv in [varInt64, varDouble, varOleStr, varBoolean]) then
      raise ESuvtaktikoLa8os.CreateFmt(sAduvatoDiabasmaEkfrasns,
        [orismata[i].KwdikasKaiPerigrafn], orismata[i].Perioxn);
    if (orismata[i].TuposDedomevwv = varBoolean) then
      raise ESuvtaktikoLa8os.Create(sAduvatoDiabasmaLogikwv, orismata[i].Perioxn);
    Inc(i);
  until (p.dk.TrSus.Tupos <> sKomma);
end;

function TEvtolnDiabase.Ektelesn: TEvtoln;
{***************************************************************************************************
               
   .  nDiadikasiaEisodou        ,
         .
      "      "
  123,    ,    123    .
***************************************************************************************************}
var
  i: integer;
begin
  result := inherited Ektelesn;                //       
  if orismata.Pln8os = 0 then                                                    //  
    p.Eisodos(nil)                                                         //   Enter
  else for i := 0 to orismata.Pln8os - 1 do
    p.Eisodos(orismata[i]);                  //   ,     
end;

procedure TEvtolnGraye.Avagvwrise;
{***************************************************************************************************
       .      .
       ,       ..
  ..., ' '      .
***************************************************************************************************}
resourcestring
  sOxiPivakasStnGraye = '     . '
    + '    -     .';
  sOxiGraceSeSuvartnseis = '       ,  '
    + '             '
    + ' .            '
    + ' .';
var
  i: integer;
begin
  if p.TY.evtoles[0] is TEvtolnSuvartnsn then
    if ru8miseis[ruOxiGraceDiabaseSeSuvartnseis].timn then
      raise ESuvtaktikoLa8os.Create(sOxiGraceSeSuvartnseis, p.dk.TrSus.Perioxn);
  i := 0;
  if p.dk.EpSus.Tupos <> sTelosGrammns then                 //  ,  Enter
  repeat
    p.dk.Epomevo;                                          //      
    orismata.Eisagwgn(Ekfrasn(p.dk));
    if (orismata[i].TuposDedomevwv and varArray) <> 0 then
      raise ESuvtaktikoLa8os.Create(sOxiPivakasStnGraye, orismata[i].Perioxn);
    Inc(i);
  until (p.dk.TrSus.Tupos <> sKomma)
  else p.dk.Epomevo;                                                    //   
end;

function TEvtolnGraye.Ektelesn: TEvtoln;
{***************************************************************************************************
             .
***************************************************************************************************}
var
  i: integer;
  keimevo: WideString;     //          
  fix: variant;
  grafnkeAri8mos: boolean;                                         //..  1, 2   1 2
begin
  result := inherited Ektelesn;                //       
  if orismata.Pln8os = 0 then                                                      //  
    p.Ejodos('')                                                              //   Enter
  else begin
    keimevo := '';
    grafnkeAri8mos := false;
    for i := 0 to orismata.Pln8os - 1 do
    begin                                    // Windows (.. 98 first edition)  
      VarClear(fix);                             //  VarCopy ( ),  
      fix := orismata[i].Timn;                         //  source    dest!!!
      if grafnkeAri8mos and ru8miseis[ruKevaMetajuAri8mwvStnGrace].timn
        and (orismata[i].TuposDedomevwv in [varInt64, varDouble]) then
        keimevo := keimevo + ' ' + VariantWsKeimevo(fix)
      else
        keimevo := keimevo + VariantWsKeimevo(fix);
      grafnkeAri8mos := orismata[i].TuposDedomevwv in [varInt64, varDouble];
    end;
    p.Ejodos(keimevo);
  end;
end;

resourcestring
  sPerimevaAllnDL = '    %s  ' + WBR + '%s';
  sXreiazetaiLogiknSuv8nkn = '    %s' + WBR
    + '   ,    .' + WBR + '%s';

procedure TEvtolnAv.Avagvwrise;
{***************************************************************************************************
       .         .
          _ ( _   
  ).
  //orismata[0] =  , orismata[1] =  _, 2..    
***************************************************************************************************}
begin
  p.dk.Epomevo;                                                          //   
  orismata.Eisagwgn(Ekfrasn(p.dk));
  if (orismata[0].TuposDedomevwv <> varBoolean) then
    raise ESuvtaktikoLa8os.CreateFmt(sXreiazetaiLogiknSuv8nkn,
      [Kwdikas, orismata[0].KwdikasKaiPerigrafn], orismata[0].Perioxn);
  if p.dk.TrSus.Tupos <> sTote then
    raise ESuvtaktikoLa8os.CreateFmt(sPerimevaAllnDL,
      [mPKTDL(sTote), p.dk.TrSus.KwdikasKaiPerigrafn], p.dk.TrSus.Perioxn);
  p.dk.Epomevo;                                                                //  
  nTeleutaiaAlliwsAv := self;                            //  _     
  p.TY.W8nsn(self);
end;

procedure TEvtolnAv.SuvdeseAlliws(nApopavwEvtoln, evtolnAlliws: TEvtoln);
{***************************************************************************************************
     _  .    
     ,        _.
          .
***************************************************************************************************}
resourcestring
  sParapavisioAlliws =
    '  %s           %s';
begin
  if (nTeleutaiaAlliwsAv = nil) then                                   //    
    raise ESuvtaktikoLa8os.CreateFmt(sParapavisioAlliws, [evtolnAlliws.Kwdikas, Kwdikas],
      evtolnAlliws.Perioxn);
  nTeleutaiaAlliwsAv.nEvtolnDiakladwsns := evtolnAlliws;
  SetLength(evtolesGiaTelosAv, Length(evtolesGiaTelosAv) + 1);
  evtolesGiaTelosAv[High(evtolesGiaTelosAv)] := nApopavwEvtoln;
  if (evtolnAlliws is TEvtolnAlliws) then                                  //   _
    nTeleutaiaAlliwsAv := nil                                     //    
  else nTeleutaiaAlliwsAv := (evtolnAlliws as TEvtolnDiakladwsns);
end;

procedure TEvtolnAv.SuvdeseTelosAv(evtolnTelosAv: TEvtoln);
{***************************************************************************************************
      .     .
       < 0 
     \\- ' '
     \   <- 1 ---------------------------
     _  < 10                    \
     \\- ''               \
     \   <- 2 ------------------------\ \
                                 \ \
      \- '  !'      \ \
         <- 3 ----------------------\ \ \
     _   <--------------------+-+-\
    _     SuvdeseAlliws      
     .            
  _ (          ,    
 _).
           _.
     evtolesGiaTelosAv    ,    " "
 ,     .
***************************************************************************************************}
var
  i: integer;
begin
  if (nTeleutaiaAlliwsAv <> nil) then
    nTeleutaiaAlliwsAv.nEvtolnDiakladwsns := evtolnTelosAv;
  for i := 0 to High(evtolesGiaTelosAv) do
    evtolesGiaTelosAv[i].nEpomevnEvtoln := evtolnTelosAv;
  evtolesGiaTelosAv := nil;     //        
end;

resourcestring
  sYparxeiAvoixtnEvtoln =
    '     %s        %s';

procedure TEvtolnAlliwsAv.Avagvwrise;
{***************************************************************************************************
       _.        
 .
***************************************************************************************************}
var
  tee: TEvtoln;                                            //    
begin
  tee := p.TY.Ejw8nsn;
  if not (tee is TEvtolnAv) then
    raise ESuvtaktikoLa8os.CreateFmt(sYparxeiAvoixtnEvtoln, [Kwdikas, tee.Kwdikas], Perioxn);
  (tee as TEvtolnAv).SuvdeseAlliws(p.TY.te, self);     //        
  p.dk.Epomevo;                                      //       
  orismata.Eisagwgn(Ekfrasn(p.dk));
  if (orismata[0].TuposDedomevwv <> varBoolean) then
    raise ESuvtaktikoLa8os.CreateFmt(sXreiazetaiLogiknSuv8nkn,
      [Kwdikas, orismata[0].KwdikasKaiPerigrafn], orismata[0].Perioxn);
  if p.dk.TrSus.Tupos <> sTote then
    raise ESuvtaktikoLa8os.CreateFmt(sPerimevaAllnDL,
      [mPKTDL(sTote), p.dk.TrSus.KwdikasKaiPerigrafn], p.dk.TrSus.Perioxn);
  p.dk.Epomevo;                                                                //  
  p.TY.W8nsn(tee);                                                    //    
end;

procedure TEvtolnAlliws.Avagvwrise;
var
  tee: TEvtoln;                                            //    
begin
  tee := p.TY.Ejw8nsn;
  if not (tee is TEvtolnAv) then
    raise ESuvtaktikoLa8os.CreateFmt(sYparxeiAvoixtnEvtoln, [Kwdikas, tee.Kwdikas], Perioxn);
  p.dk.Epomevo;                                                             //  
  (tee as TEvtolnAv).SuvdeseAlliws(p.TY.te, self);
  p.TY.W8nsn(tee);                                                    //    
end;

procedure TEvtolnTelosAv.Avagvwrise;
var
  tee: TEvtoln;                                            //    
begin
  tee := p.TY.Ejw8nsn;                                //       
  if not (tee is TEvtolnAv) then
    raise ESuvtaktikoLa8os.CreateFmt(sYparxeiAvoixtnEvtoln, [Kwdikas, tee.Kwdikas], Perioxn);
  p.dk.Epomevo;                                                           //  _
  (tee as TEvtolnAv).SuvdeseTelosAv(self);
end;

procedure TEvtolnEpileje.Avagvwrise;
{***************************************************************************************************
       .       
 .          _ (  
     ).
             .
      :
     + ' ' + 
    '' .. '', c, '' + c:                 //    
                                                              //  
    :                                //      _: ;
     
   _
         ,      , 
      >= 1   <= 2   ,  
    = ޻.

         :
 1) :      ,      .  
       ,           
    ,     .          nil, 
          .      nil,   
     .
         ,       
 2) ) :  subclassing     TEkfrasn.Timn     , 
        .   Create      
           ,        
     override!
    1)           .    create
        (  )         
     .       .
    2)            TEkfrasn   
        ,          
             . ?????????  ??????????
      2    ,   .
***************************************************************************************************}
resourcestring
  sOxiPivakasStnvEpileje = '        ,'
     + '   .' + WBR + '%s';
begin
  p.dk.Epomevo;                                                     //   
  orismata.Eisagwgn(Ekfrasn(p.dk));                        //   
  if (orismata[0].TuposDedomevwv and varArray) <> 0 then                         //  
    raise ESuvtaktikoLa8os.CreateFmt(sOxiPivakasStnvEpileje, [orismata[0].KwdikasKaiPerigrafn],
      orismata[0].Perioxn);
  p.TY.W8nsn(self);
end;

procedure TEvtolnEpileje.SuvdesePeriptwsn(nApopavwEvtoln, evtolnPeriptwsn: TEvtoln);
{***************************************************************************************************
        .     
    ,        _.
            .
***************************************************************************************************}
resourcestring
  sParapavisiaPeriptwsn = '  %s        '
   + '    %s';
begin
  if nTeleutaiaPeriptwsn is TEvtolnPeriptwsnAlliws then   // nTeleutaiaPeriptwsn = nil  false
    raise ESuvtaktikoLa8os.CreateFmt(sParapavisiaPeriptwsn, [evtolnPeriptwsn.Kwdikas, Kwdikas],
      evtolnPeriptwsn.Perioxn);
  if nTeleutaiaPeriptwsn <> nil then                       //      
  begin
    (nTeleutaiaPeriptwsn as TEvtolnPeriptwsn).nEvtolnDiakladwsns := evtolnPeriptwsn;
    SetLength(evtolesGiaTelosEpilogwv, Length(evtolesGiaTelosEpilogwv) + 1);
    evtolesGiaTelosEpilogwv[High(evtolesGiaTelosEpilogwv)] := nApopavwEvtoln;
  end;
  nTeleutaiaPeriptwsn := evtolnPeriptwsn;
end;

function TEvtolnEpileje.Ektelesn: TEvtoln;
begin
  result := inherited Ektelesn;                //       
  apotimnsn := orismata[0].Timn;                           //       
end;

procedure TEvtolnEpileje.SuvdeseTelosEpilogwv(evtolnTelosEpilogwv: TEvtoln);
{***************************************************************************************************
   :   '             
   true,     .  (if nSuv8nkn = nil)      
       .           
   .
    
    ..0
   | \- '   !' -|
    0..9                   |
   | \- '' -----------| |
    10..16               | |
   | \- ' ' ------| | |
    16..20             | | |
   | \- '!' -------| | | |
               | | | |
     \- ' ' --| | | | |
   _     <-----+-+-+-+-|
       SuvdesePeriptwsn      
     .      .

           _.
     evtolesGiaTelosEpilogwv    ,    "
 " ,     .
***************************************************************************************************}
var
  i: integer;
begin
  Assert(nTeleutaiaPeriptwsn <> nil);
  if nTeleutaiaPeriptwsn is TEvtolnPeriptwsn then     //    TEvtolnPeriptwsnAlliws
    (nTeleutaiaPeriptwsn as TEvtolnPeriptwsn).nEvtolnDiakladwsns := evtolnTelosEpilogwv;
  for i := 0 to High(evtolesGiaTelosEpilogwv) do
    evtolesGiaTelosEpilogwv[i].nEpomevnEvtoln := evtolnTelosEpilogwv;
  evtolesGiaTelosEpilogwv:= nil;//        
end;

procedure TEvtolnPeriptwsn.ElegjeEkfrasn(ekf: TSustatiko);
{***************************************************************************************************
              ,
        .
  prajeis[High(prajeis)]       .
***************************************************************************************************}
resourcestring
  sMovoSta8eresStnvPeriptwsn = '        ,'
     + '        .' + WBR + '%s';
  sAsumbatnSugkrisnStnvPeriptwsn = '        ,' + WBR
   + '%s' + WBR + '   .' + WBR + '%s';
  sOxiAvisotntesLogikwvStnvPeriptwsn = '      '
     + '(<, <=, >, >=, ..)          (=, <>).';
begin
  if not ekf.Ypostnrizei(lsSta8ernTimn) or ((ekf.TuposDedomevwv and varArray) <> 0) then
    raise ESuvtaktikoLa8os.CreateFmt(sMovoSta8eresStnvPeriptwsn, [ekf.KwdikasKaiPerigrafn],
      ekf.Perioxn);
  if (nEvtolnEpileje.orismata[0].TuposDedomevwv <> ekf.TuposDedomevwv) and
   not ((nEvtolnEpileje.orismata[0].TuposDedomevwv in [varInt64, varDouble]) and
        (ekf.TuposDedomevwv in [varInt64, varDouble])) then
    raise ESuvtaktikoLa8os.CreateFmt(sAsumbatnSugkrisnStnvPeriptwsn,
      [nEvtolnEpileje.orismata[0].KwdikasKaiPerigrafn, ekf.KwdikasKaiPerigrafn], ekf.Perioxn)
  else if (ekf.TuposDedomevwv = varBoolean) and
   not (prajeis[High(prajeis)] in [sIsov] + sDiaforo) then
    raise ESuvtaktikoLa8os.Create(sOxiAvisotntesLogikwvStnvPeriptwsn, ekf.Perioxn);
end;

procedure TEvtolnPeriptwsn.Prospa8nseApotimnsn;
{***************************************************************************************************
    inherited TEvtolnDiakladwsns.Ektelesn         .
     (  ),     try   
 override  KaveApotimnsn.
***************************************************************************************************}
var
  i: integer;
  trexousa: boolean;
  timn1, timn2: Variant;
begin
  fApotimnsn := false;
  if nEvtolnEpileje.TuposDedomevwv = varDouble then
    timn1 := RoundTo(nEvtolnEpileje.apotimnsn, -ru8miseis[ruSnmavtikaDekadika].timn)
  else timn1 := nEvtolnEpileje.apotimnsn;
  i := 0;
  while i < orismata.Pln8os do
  begin
    if orismata[i].TuposDedomevwv = varDouble then
      timn2 := RoundTo(orismata[i].Timn, -ru8miseis[ruSnmavtikaDekadika].timn)
    else timn2 := orismata[i].Timn;
    case prajeis[i] of
      sMikrotero:                                  trexousa := timn1 < timn2;
      sIsov:                                       trexousa := timn1 = timn2;
      sMegalutero:                                 trexousa := timn1 > timn2;
      sDiaforoANSI, sDiaforoUNICODE:               trexousa := timn1 <> timn2;
      sMikroteroHIsoANSI, sMikroteroHIsoUNICODE:   trexousa := timn1 <= timn2;
      sMegaluteroHIsoANSI, sMegaluteroHIsoUNICODE: trexousa := timn1 >= timn2;
    else                // sDiastnma   ,       compiler...
      begin
        Inc(i);                                               //     
        trexousa := (timn1 >= timn2) and (timn1 <= orismata[i].Timn);
      end;
    end;
    fApotimnsn := fApotimnsn or trexousa;             //     ,  
    Inc(i);
  end;
end;

procedure TEvtolnPeriptwsn.SarwseEkfraseis;
{***************************************************************************************************
    Avagvwrise     .     .
***************************************************************************************************}
begin
  repeat
    p.dk.Epomevo;                                             //     
    SetLength(prajeis, Length(prajeis) + 1);
    if p.dk.TrSus.Tupos in
     [sMikrotero, sIsov, sMegalutero] + sDiaforo + sMikroteroHIso + sMegaluteroHIso then
    begin
      prajeis[High(prajeis)] := p.dk.TrSus.Tupos;
      p.dk.Epomevo;                                              //  <, <>, <=, =, >  >=
      orismata.Eisagwgn(Ekfrasn(p.dk));            //       <, > 
    end
    else begin             //          , .. 9  1..3
      orismata.Eisagwgn(Ekfrasn(p.dk));
      if p.dk.TrSus.Tupos <> sDiastnma then                                           // 
        prajeis[High(prajeis)] := sIsov
      else begin
        prajeis[High(prajeis)] := p.dk.TrSus.Tupos;                                      //sDiastnma
        ElegjeEkfrasn(orismata[orismata.Pln8os - 1]);        //   
        SetLength(prajeis, Length(prajeis) + 1);
        prajeis[High(prajeis)] := p.dk.TrSus.Tupos;        //   "",  
        p.dk.Epomevo;                                                        //  
        orismata.Eisagwgn(Ekfrasn(p.dk));
      end;
    end;
    ElegjeEkfrasn(orismata[orismata.Pln8os - 1]);       //      
  until p.dk.TrSus.Tupos <> sKomma;
end;

procedure TEvtolnPeriptwsn.Avagvwrise;
{***************************************************************************************************
        KAI      .
             .
           .
***************************************************************************************************}
begin
  if not (p.TY.Krufokoitagma is TEvtolnEpileje) then
    raise ESuvtaktikoLa8os.CreateFmt(sYparxeiAvoixtnEvtoln, [Kwdikas, p.TY.Krufokoitagma.Kwdikas],
      Perioxn);
  nEvtolnEpileje := (p.TY.Krufokoitagma as TEvtolnEpileje);
  nEvtolnEpileje.SuvdesePeriptwsn(p.TY.te, self);
//          "   ".
  SarwseEkfraseis;
end;

procedure TEvtolnPeriptwsnAlliws.Avagvwrise;
var
  tee: TEvtoln;                                            //    
begin
  p.dk.Epomevo;                                                           //  
  tee := p.TY.Krufokoitagma;
  if not (tee is TEvtolnEpileje) then
    raise ESuvtaktikoLa8os.CreateFmt(sYparxeiAvoixtnEvtoln, [Kwdikas, tee.Kwdikas], Perioxn);
  (tee as TEvtolnEpileje).SuvdesePeriptwsn(p.TY.te, self);
//          "   ".
  p.dk.Epomevo;                                                              //  
end;

procedure TEvtolnTelosEpilogwv.Avagvwrise;
var
  tee: TEvtoln;                                            //    
begin
  tee := p.TY.Ejw8nsn;                          //        
  if not (tee is TEvtolnEpileje) then
    raise ESuvtaktikoLa8os.CreateFmt(sYparxeiAvoixtnEvtoln, [Kwdikas, tee.Kwdikas], Perioxn);
  p.dk.Epomevo;                                                     //  _
  (tee as TEvtolnEpileje).SuvdeseTelosEpilogwv(self);
end;

procedure TEvtolnOso.Avagvwrise;
{***************************************************************************************************
      .           
 .
***************************************************************************************************}
begin
  p.dk.Epomevo;                                                         //   
  orismata.Eisagwgn(Ekfrasn(p.dk));
  if orismata[0].TuposDedomevwv <> varBoolean then
    raise ESuvtaktikoLa8os.CreateFmt(sXreiazetaiLogiknSuv8nkn,
      [Kwdikas, orismata[0].KwdikasKaiPerigrafn], orismata[0].Perioxn);
  if p.dk.TrSus.Tupos <> sEpavalabe then
    raise ESuvtaktikoLa8os.CreateFmt(sPerimevaAllnDL,
      [mPKTDL(sEpavalabe), p.dk.TrSus.KwdikasKaiPerigrafn], p.dk.TrSus.Perioxn);
  p.dk.Epomevo;                                                           //  
  p.TY.W8nsn(self);
end;

procedure TEvtolnMexriMeBnma.Prospa8nseApotimnsn;
{***************************************************************************************************
          _    .
***************************************************************************************************}
begin
  if orismata[0].Ypostnrizei(lsSta8ernApotimnsn) then
    if (orismata.Pln8os <= 1) or orismata[1].Ypostnrizei(lsSta8ernApotimnsn) then
    try
      KaveApotimnsn;
    finally
    end;
end;

procedure TEvtolnMexriMeBnma.KaveApotimnsn;
{***************************************************************************************************
  KaveApotimnsn       MexriMeBnma.   
      (timnMexri = UnAssigned),    ,   
 ' ,        .
***************************************************************************************************}
resourcestring
  sBnma0StnGia = '     ,        '
   + ' ';
var
  snmavtika: double;
begin
  snmavtika := Power(0.1, ru8miseis[ruSnmavtikaDekadika].timn);
  if VarIsEmpty(timnMexri) then                                                       //
  begin
    timnMexri := orismata[0].Timn;
    if orismata.Pln8os <= 1 then timnBnmatos := 1
    else timnBnmatos := orismata[1].Timn;
    if timnBnmatos = 0 then
      raise ELa8osEktelesns.Create(sBnma0StnGia, orismata[1].Perioxn)
    else if timnBnmatos > 0 then prajn := sMikroteroHIsoANSI
    else prajn := sMegaluteroHIsoANSI;
  end
    else nEvtolnGia.orismata[0].Timn := nEvtolnGia.orismata[0].Timn + timnBnmatos;
  if (nEvtolnGia.orismata[0].TuposDedomevwv = varDouble) or (VarType(timnMexri) = varDouble) then
    if prajn in sMikroteroHIso then
      fApotimnsn := nEvtolnGia.orismata[0].Timn <= timnMexri + snmavtika
    else fApotimnsn := nEvtolnGia.orismata[0].Timn >= timnMexri - snmavtika
  else
    if prajn in sMikroteroHIso then
      fApotimnsn := nEvtolnGia.orismata[0].Timn <= timnMexri
    else fApotimnsn := nEvtolnGia.orismata[0].Timn >= timnMexri;
//todo:    prajn  prosnmo     ...    
end;

procedure TEvtolnMexriMeBnma.Avagvwrise;
{***************************************************************************************************
   TrSus = .
***************************************************************************************************}
resourcestring
  sEkfrasnMexriOxiAri8mntikn =
    '        .' + WBR + '%s';
  sEkfrasnMeBnmaOxiAri8mntikn =
    '  _      .' + WBR + '%s';
  sMetrntnsAsumbatosMeBnma = '         '
    + '       .';
  sDevEpitrepetaiMe = '           '
    + '             '
    + ',         .';
var
  brnkaMeBnma: boolean;
begin
  nEvtolnGia := p.TY.Krufokoitagma as TEvtolnGia;                 //      
  p.dk.Epomevo;                                                               //  
  orismata.Eisagwgn(Ekfrasn(p.dk));
  if not (orismata[0].TuposDedomevwv in [varInt64, varDouble]) then
    raise ESuvtaktikoLa8os.CreateFmt(sEkfrasnMexriOxiAri8mntikn, [orismata[0].KwdikasKaiPerigrafn],
      orismata[0].Perioxn);
//      
  brnkaMeBnma := true;
  if p.dk.TrSus.Tupos = sMe then
    if not ru8miseis[ruMeBnma].timn then
      raise ESuvtaktikoLa8os.Create(sDevEpitrepetaiMe, p.dk.TrSus.Perioxn)
    else begin
      p.dk.Epomevo;
      if p.dk.TrSus.Tupos <> sBnma then
        raise ESuvtaktikoLa8os.CreateFmt(sPerimevaAllnDL,
          [mPKTDL(sBnma), p.dk.TrSus.KwdikasKaiPerigrafn], p.dk.TrSus.Perioxn);
    end
  else if p.dk.TrSus.Tupos <> sMe_Bnma then
    brnkaMeBnma := false;
  if brnkaMeBnma then
  begin
    p.dk.Epomevo;                                                //  "",  "_"
    orismata.Eisagwgn(Ekfrasn(p.dk));
    if not (orismata[1].TuposDedomevwv in [varInt64, varDouble]) then
      raise ESuvtaktikoLa8os.CreateFmt(sEkfrasnMeBnmaOxiAri8mntikn,
        [orismata[1].KwdikasKaiPerigrafn], orismata[0].Perioxn);
    if (nEvtolnGia.orismata[0].TuposDedomevwv = varInt64)
     and (orismata[1].TuposDedomevwv = varDouble) then
      raise ESuvtaktikoLa8os.Create(sMetrntnsAsumbatosMeBnma, orismata[1].Perioxn);
  end;
end;

procedure TEvtolnGia.Avagvwrise;
{***************************************************************************************************
      .     ,  ,   ,
       TEvtolnMexriMeBnma     ,  
           .
            :
  #  Basic,  Pascal   Delphi          ( 
      Basic),   C    .
  #       Pascal      ,   Delphi, 
    Basic (       C)       .
  #  Delphi   Pascal       ,      -
      .  Basic   C         .
       Basic,        :
      <- _
     _ <- _
     _ <- _
      ( [>=  <=] _) 
       
        <-  + _
     _
***************************************************************************************************}
resourcestring
  sMetrntnsOxiMetablntn = '        .' + WBR + '%s';
  sMetrntnsOxiAri8mntikos = '        .' + WBR + '%s';
  sEkfrasnApoOxiAri8mntikn = '      .' + WBR + '%s';
  sApoAsumbatoMeMetrntn = '    ,     .'
    + WBR + '%s';
begin
  p.dk.Epomevo;                                                                 //  
  orismata.Eisagwgn(Ekfrasn(p.dk));
  if not orismata[0].Ypostnrizei(lsQeseTimn) then
    raise ESuvtaktikoLa8os.CreateFmt(sMetrntnsOxiMetablntn,
      [orismata[0].KwdikasKaiPerigrafn], orismata[0].Perioxn)
  else if not (orismata[0].TuposDedomevwv in [varInt64, varDouble]) then
    raise ESuvtaktikoLa8os.CreateFmt(sMetrntnsOxiAri8mntikos,
      [orismata[0].KwdikasKaiPerigrafn], orismata[0].Perioxn);
  if p.dk.TrSus.Tupos <> sApo then
    raise ESuvtaktikoLa8os.CreateFmt(sPerimevaAllnDL,
      [mPKTDL(sApo), p.dk.TrSus.KwdikasKaiPerigrafn], p.dk.TrSus.Perioxn);
  p.dk.Epomevo;                                                                 //  
  orismata.Eisagwgn(Ekfrasn(p.dk));
  if not (orismata[1].TuposDedomevwv in [varInt64, varDouble]) then
    raise ESuvtaktikoLa8os.CreateFmt(sEkfrasnApoOxiAri8mntikn,
      [orismata[1].KwdikasKaiPerigrafn], orismata[1].Perioxn)
  else if (orismata[0].TuposDedomevwv = varInt64) and (orismata[1].TuposDedomevwv = varDouble) then
    raise ESuvtaktikoLa8os.CreateFmt(sApoAsumbatoMeMetrntn,
      [orismata[1].KwdikasKaiPerigrafn], orismata[1].Perioxn);
  if p.dk.TrSus.Tupos <> sMexri then
    raise ESuvtaktikoLa8os.CreateFmt(sPerimevaAllnDL,
      [mPKTDL(sMexri), p.dk.TrSus.KwdikasKaiPerigrafn], p.dk.TrSus.Perioxn);
  p.TY.W8nsn(self);               //        nEvtolnMexriMeBnma
  nEvtolnMexriMeBnma := TEvtolnMexriMeBnma.Create(p);              //   
// Create  TEvtolnMexriMeBnma     ,       
end;

function TEvtolnGia.Ektelesn: TEvtoln;
begin
  result := inherited Ektelesn;                //       
  orismata[0].Timn := orismata[1].Timn;                                  // <-  
  VarClear(nEvtolnMexriMeBnma.timnMexri);        //      
end;

procedure TEvtolnTelosEpavalnyns.Avagvwrise;
var
  tee: TEvtoln;                                            //    
begin
  tee := p.TY.Ejw8nsn;                                               //     
  if (tee is TEvtolnOso) then
  begin
    (tee as TEvtolnOso).nEvtolnDiakladwsns := self;
    p.TY.te.nEpomevnEvtoln := tee;              //        
  end
  else if (tee is TEvtolnGia) then
  begin
    (tee as TEvtolnGia).nEvtolnMexriMeBnma.nEvtolnDiakladwsns := self;
    p.TY.te.nEpomevnEvtoln := (tee as TEvtolnGia).nEvtolnMexriMeBnma;  //   
  end
  else
    raise ESuvtaktikoLa8os.CreateFmt(sYparxeiAvoixtnEvtoln, [Kwdikas, tee.Kwdikas], Perioxn);
  p.dk.Epomevo;                                                    //  _
end;

procedure TEvtolnArxnEpavalnyns.Avagvwrise;
{***************************************************************************************************
   TrSus = _.
***************************************************************************************************}
begin
  p.dk.Epomevo;                                             //   _
  p.TY.W8nsn(self);
end;

procedure TEvtolnMexrisOtou.Avagvwrise;
{***************************************************************************************************
   TrSus = _.           
 _.
***************************************************************************************************}
var
  tee: TEvtoln;                                            //    
begin
  tee := p.TY.Ejw8nsn;
  if not (tee is TEvtolnArxnEpavalnyns) then
    raise ESuvtaktikoLa8os.CreateFmt(sYparxeiAvoixtnEvtoln, [Kwdikas, tee.Kwdikas], Perioxn);
  p.dk.Epomevo;                                                 //   _
  orismata.Eisagwgn(Ekfrasn(p.dk));
  if orismata[0].TuposDedomevwv <> varBoolean then
    raise ESuvtaktikoLa8os.CreateFmt(sXreiazetaiLogiknSuv8nkn,
      [Kwdikas, orismata[0].KwdikasKaiPerigrafn], orismata[0].Perioxn);
  nEvtolnDiakladwsns := tee.nEpomevnEvtoln;
end;

{ TEvtolnKalese }

procedure TEvtolnKalese.Avagvwrise;
{***************************************************************************************************
   TrSus = .
***************************************************************************************************}
resourcestring
  sOxiKaleseSuvartnsn = '   ,    '
    + '   .';
  sPerimevaOvomaDiadikasias = '  .';
begin
  p.dk.Epomevo;                                                              //  
  if p.dk.TrSus is TEkfrasnAvagvwristikou then
    if (p.dk.TrSus as TEkfrasnAvagvwristikou).avagvwristiko = nil then
      raise ESuvtaktikoLa8os.CreateFmt(sAgvwstoAvagvwristiko, [p.dk.TrSus.Kwdikas],
        p.dk.TrSus.Perioxn)
    else if (p.dk.TrSus as TEkfrasnAvagvwristikou).avagvwristiko is TSuvartnsn then
      raise ESuvtaktikoLa8os.Create(sOxiKaleseSuvartnsn, p.dk.TrSus.Perioxn)
    else if (p.dk.TrSus as TEkfrasnAvagvwristikou).avagvwristiko is TKuriwsProgramma then
      orismata.Eisagwgn(TEkfrasnKuriwsProgrammatos.Create(p.dk))
    else if (p.dk.TrSus as TEkfrasnAvagvwristikou).avagvwristiko is TDiadikasia then
      orismata.Eisagwgn(TEkfrasnDiadikasias.Create(p.dk))
    else raise ESuvtaktikoLa8os.Create(sPerimevaOvomaDiadikasias, p.dk.TrSus.Perioxn)
  else raise ESuvtaktikoLa8os.Create(sPerimevaOvomaDiadikasias, p.dk.TrSus.Perioxn);      //, 
end;

function TEvtolnKalese.Ektelesn: TEvtoln;
var
  i: integer;
  v: Variant;
begin
  result := inherited Ektelesn;                //       
  with orismata[0] as TEkfrasnYpoprogrammatos do for i := 0 to orismata.Pln8os - 1 do
//  UnAssigned      
    if (orismata[i] is TEkfrasnAvagvwristikou)
     and ((orismata[i] as TEkfrasnAvagvwristikou).avagvwristiko is TMetablntn) then
      p.stoiba.Push(p.stoiba[
       ((orismata[i] as TEkfrasnAvagvwristikou).avagvwristiko as TMetablntn).Metatopisn]^)
    else if (orismata[i] is TEkfrasnStoixeiouPivaka) then  
      upoprogramma.p.stoiba.Push((orismata[i] as TEkfrasnStoixeiouPivaka).DesTimnHUnassigned)
    else
      p.stoiba.Push(orismata[i].Timn);
  (orismata[0] as TEkfrasnYpoprogrammatos).upoprogramma.Klnsn; //     !
//todo:               , ..
// result := inherited Ektelesn;
//,         TelosDiadikasias;
  for i := (orismata[0] as TEkfrasnYpoprogrammatos).orismata.Pln8os - 1 downto 0 do
    if (orismata[0] as TEkfrasnYpoprogrammatos).orismata[i].Ypostnrizei(lsQeseTimn) then
    begin
      v := p.stoiba.Pop;
      if VarIsEmpty(v) then
        (orismata[0] as TEkfrasnYpoprogrammatos).orismata[i].Timn := Unassigned        //todo: fixme
      else
        (orismata[0] as TEkfrasnYpoprogrammatos).orismata[i].Timn := v;
    end
    else p.stoiba.Pop;                      //    ,   
end;

{ TEvtolnProgramma }

procedure TEvtolnProgramma.Avagvwrise;
{***************************************************************************************************
   TrSus = .      .
***************************************************************************************************}
resourcestring
  sPerimevaOvomaProgrammatos = '    .';
  sAkuroOvomaProgrammatos = '   : %s.';
begin
  p.fKP := p.TY as TKuriwsProgramma;
  if p.dk.Epomevo is TTelosGrammns then
    raise ESuvtaktikoLa8os.Create(sPerimevaOvomaProgrammatos, p.dk.TrSus.Perioxn)
  else if not (p.dk.TrSus is TEkfrasnAvagvwristikou) then
    raise ESuvtaktikoLa8os.CreateFmt(sAkuroOvomaProgrammatos, [p.dk.TrSus.Kwdikas],
      p.dk.TrSus.Perioxn);
  p.TY.ovoma := p.dk.TrSus;
  p.dk.Epomevo;
  p.TY.W8nsn(self);
end;

procedure TEvtolnTelosProgrammatos.Avagvwrise;
{***************************************************************************************************
   TrSus = _.
***************************************************************************************************}
resourcestring
  sPerimevaOvomaKuriwsProgrammatos = '      %s' + WBR + '%s';
var
  tee: TEvtoln;
begin
  tee := p.TY.Ejw8nsn;
  if tee.ClassType <> TEvtolnProgramma then
    raise ESuvtaktikoLa8os.CreateFmt(sYparxeiAvoixtnEvtoln, [Kwdikas, tee.Kwdikas], Perioxn);
  p.dk.Epomevo;                                                  //  _
  if (p.dk.TrSus is TEkfrasnAvagvwristikou)                            //  ,
   and ((p.dk.TrSus as TEkfrasnAvagvwristikou).avagvwristiko = p.TY) then   //   
     p.dk.Epomevo                                                 //,   
  else if not (p.dk.TrSus is TTelosGrammns) then      //      Enter
    raise ESuvtaktikoLa8os.CreateFmt(sPerimevaOvomaKuriwsProgrammatos,
      [p.TY.ovoma.Kwdikas, p.dk.TrSus.KwdikasKaiPerigrafn], p.dk.TrSus.Perioxn);
end;

procedure TEvtolnDiadikasia.Avagvwrise;
{***************************************************************************************************
   TrSus =   .      ,
           .
  TEvtolnDiadikasia  private  protected   TDiadikasia.   
 ,         ...
***************************************************************************************************}
resourcestring
  sAkuroOvomaYpoprogrammatos = '   : %s';
  sPerimevaOvomaParametrou = '  .' + WBR + '%s';
  sOxiParametroiMeToIdioOvoma = '       .';
  sPerimevaKommaHParev8esn = '    ';
var
  i: integer;
begin
  p.TY.W8nsn(self);                         //           Exit
  if not (p.dk.Epomevo is TEkfrasnAvagvwristikou) then
    raise ESuvtaktikoLa8os.CreateFmt(sAkuroOvomaYpoprogrammatos, [p.dk.TrSus.Kwdikas],
      p.dk.TrSus.Perioxn);
  p.TY.EgkuroOvoma(p.dk.TrSus as TEkfrasnAvagvwristikou);//,  p.TY.ovoma   
  p.TY.ovoma := p.dk.TrSus;
  if p.dk.Epomevo.Tupos <> sArParev8esn then Exit;
  with p.TY as TDiadikasia do
  repeat
    if not (p.dk.Epomevo is TEkfrasnAvagvwristikou) then              //  (   
      raise ESuvtaktikoLa8os.CreateFmt(sPerimevaOvomaParametrou,
        [p.dk.TrSus.KwdikasKaiPerigrafn], p.dk.TrSus.Perioxn);
    EgkuroOvoma(p.dk.TrSus as TEkfrasnAvagvwristikou);
    SetLength(parametroi, Length(parametroi) + 1);
    parametroi[High(parametroi)] := p.dk.TrSus as TEkfrasnAvagvwristikou;
    for i := 0 to High(parametroi) - 1 do                        //    
      if KefalaioAtovo(parametroi[i].Kwdikas) = KefalaioAtovo(parametroi[High(parametroi)].Kwdikas)
//todo:  KefalaioAtovo(...) = KefalaioAtovo(...) ,   case insensitive compare
        then raise ESuvtaktikoLa8os.Create(sOxiParametroiMeToIdioOvoma,
          parametroi[High(parametroi)].Perioxn);
  until (p.dk.Epomevo.Tupos <> sKomma);
  if p.dk.TrSus.Tupos <> sDeParev8esn then
    raise ESuvtaktikoLa8os.Create(sPerimevaKommaHParev8esn, p.dk.TrSus.Perioxn);
  p.dk.Epomevo;
end;

procedure TEvtolnTelosDiadikasias.Avagvwrise;
{***************************************************************************************************
   TrSus = _.
***************************************************************************************************}
resourcestring
  sPerimevaToOvomaDiadikasias = '     %s.' + WBR + '%s';
var
  tee: TEvtoln;
begin
  tee := p.TY.Ejw8nsn;
  if tee.ClassType <> TEvtolnDiadikasia then                  //thanks ptsiotakis for the bug report
    raise ESuvtaktikoLa8os.CreateFmt(sYparxeiAvoixtnEvtoln, [Kwdikas, tee.Kwdikas], Perioxn);
  p.dk.Epomevo;                                                   //  _
  if (p.dk.TrSus is TEkfrasnAvagvwristikou)                            //  ,
   and ((p.dk.TrSus as TEkfrasnAvagvwristikou).avagvwristiko <> nil)                  // ,
   and ((p.dk.TrSus as TEkfrasnAvagvwristikou).avagvwristiko.ovoma = p.TY.ovoma) then
     p.dk.Epomevo            //       ,   
  else if not (p.dk.TrSus is TTelosGrammns) then      //      Enter
    raise ESuvtaktikoLa8os.CreateFmt(sPerimevaToOvomaDiadikasias,
      [p.TY.ovoma.Kwdikas, p.dk.TrSus.KwdikasKaiPerigrafn], p.dk.TrSus.Perioxn);
end;

procedure TEvtolnSuvartnsn.Avagvwrise;
{***************************************************************************************************
   TrSus = .   inherited        
    .
  TEvtolnSuvartnsn  private  protected   TSuvartnsn.   
 ,         ...
***************************************************************************************************}
resourcestring
  sPerimevaAvwKatwTeleia = '       '
     + '     :.';
  sPerimevaTonTupoSuvartnsns = '     , '
     + ' , ޻,   ޻';
begin
  inherited;                                                                       //  
  if p.dk.TrSus.Tupos <> sAvwKatwTeleia then
    raise ESuvtaktikoLa8os.Create(sPerimevaAvwKatwTeleia, p.dk.TrSus.Perioxn);
  case p.dk.Epomevo.Tupos of
    sAkeraia:    (p.TY as TSuvartnsn).fTuposDedomevwv := varInt64;
    sPragmatikn: (p.TY as TSuvartnsn).fTuposDedomevwv := varDouble;
    sXaraktnras: (p.TY as TSuvartnsn).fTuposDedomevwv := varOleStr;
    sLogikn:     (p.TY as TSuvartnsn).fTuposDedomevwv := varBoolean;
  else
    raise ESuvtaktikoLa8os.Create(sPerimevaTonTupoSuvartnsns, p.dk.TrSus.Perioxn);
  end;
  p.dk.Epomevo;                                                       //   
//  W8nsn,   inherited
end;

procedure TEvtolnTelosSuvartnsns.Avagvwrise;
{***************************************************************************************************
   TrSus = _.
***************************************************************************************************}
resourcestring
  sPerimevaToOvomaSuvartnsns = '     %s' + WBR + '%s';
var
  tee: TEvtoln;
begin
  tee := p.TY.Ejw8nsn;
  if tee.ClassType <> TEvtolnSuvartnsn then
    raise ESuvtaktikoLa8os.CreateFmt(sYparxeiAvoixtnEvtoln, [Kwdikas, tee.Kwdikas], Perioxn);
  p.dk.Epomevo;                                                    //  _
  if (p.dk.TrSus is TEkfrasnAvagvwristikou)                            //  ,
   and ((p.dk.TrSus as TEkfrasnAvagvwristikou).avagvwristiko <> nil)                  // ,
   and ((p.dk.TrSus as TEkfrasnAvagvwristikou).avagvwristiko.ovoma = p.TY.ovoma) then
     p.dk.Epomevo            //       ,   
  else if not (p.dk.TrSus is TTelosGrammns) then      //      Enter
    raise ESuvtaktikoLa8os.CreateFmt(sPerimevaToOvomaSuvartnsns,
      [p.TY.ovoma.Kwdikas, p.dk.TrSus.KwdikasKaiPerigrafn], p.dk.TrSus.Perioxn);
end;

procedure TEvtolnDnlwseis.Avagvwrise;
{***************************************************************************************************
 , , ,  .
  TEvtolnDnlwseis  private  protected   TYpoprogramma.   
 ,      ...
***************************************************************************************************}
resourcestring
  sOxiSta8eraHParametros = '         %s.';
  sPerimevaSta8eresMetablntesHArxh = '      ,'
     + '   ޻   %s.';
  sPerimevaTupoDedomevwv = '   ,     '
     + ' , ,   .';
  sPerimevaAvwKatwTeleia = '     :';
  sAkuroOvomaMetablntns = '  %s     .';
  sPerimevaMetablntesHArxh = '     '
     + '   ޻   %s.';
var
  tupos: TVarType;
  i, met, arMetablntwv: integer;
  av: TOvomasmevnSta8era;
begin
  if p.dk.TrSus.Tupos = sSta8eres then
  begin
    p.dk.Epomevo;
    ProsperaseTelosGrammns;
    while p.dk.TrSus is TEkfrasnAvagvwristikou do
    begin
      av := TOvomasmevnSta8era.Create(p);                    //      
      p.TY.EisagwgnAvagvwristikou(av);                //       FreeAndNil
      for i := 0 to High(p.TY.parametroi) do with p.TY do
        if KefalaioAtovo(av.ovoma.Kwdikas) = KefalaioAtovo(parametroi[i].Kwdikas) then
          raise ESuvtaktikoLa8os.CreateFmt(sOxiSta8eraHParametros,
            [parametroi[i].Kwdikas], av.Perioxn);
      ProsperaseTelosGrammns;
    end;
  end
  else if not (p.dk.TrSus.Tupos in [sMetablntes, sArxn]) then
    raise ESuvtaktikoLa8os.CreateFmt(sPerimevaSta8eresMetablntesHArxh,
      [p.dk.TrSus.Kwdikas], p.dk.TrSus.Perioxn);
  if p.dk.TrSus.Tupos = sMetablntes then
  begin
    p.dk.Epomevo;                                                        //  
    ProsperaseTelosGrammns;
    if not (p.dk.TrSus.Tupos in [sAkeraies, sPragmatikes, sXaraktnres, sLogikes]) then
      raise ESuvtaktikoLa8os.Create(sPerimevaTupoDedomevwv, p.dk.TrSus.Perioxn);
    arMetablntwv := 0;
    repeat
      case p.dk.TrSus.Tupos of
        sAkeraies:    tupos := varInt64;
        sPragmatikes: tupos := varDouble;
        sXaraktnres:  tupos := varOleStr;
        else tupos := varBoolean;                            //(= sLogikes),    warning
      end;
      if (p.dk.Epomevo.Tupos <> sAvwKatwTeleia) then            //   .. 
        raise ESuvtaktikoLa8os.Create(sPerimevaAvwKatwTeleia, p.dk.TrSus.Perioxn);
      repeat
        if not (p.dk.Epomevo is TEkfrasnAvagvwristikou) then          //  :   
          raise ESuvtaktikoLa8os.CreateFmt(sAkuroOvomaMetablntns, [p.dk.TrSus.Kwdikas],
            p.dk.TrSus.Perioxn);
        i := 0; //   ,        
        while (i <= High(p.TY.parametroi))
         and (KefalaioAtovo(p.dk.TrSus.Kwdikas) <> KefalaioAtovo(p.TY.parametroi[i].Kwdikas)) do
          Inc(i);
        if i <= High(p.TY.parametroi) then
          met := i - High(p.TY.parametroi) - 1                 // -High(p.TY.parametroi)  -1
        else begin
          met := arMetablntwv;
          Inc(arMetablntwv);
        end;
        if p.dk.EpSus.Tupos <> sArAgkuln then
          p.TY.EisagwgnAvagvwristikou(TMetablntn.Create(p, tupos, met))
        else p.TY.EisagwgnAvagvwristikou(TMetablntnPivakas.Create(p, tupos, met));
      until (p.dk.TrSus.Tupos <> sKomma);                        //    
      ProsperaseTelosGrammns;
    until not (p.dk.TrSus.Tupos in [sAkeraies, sPragmatikes, sXaraktnres, sLogikes]);
  end
  else if p.dk.TrSus.Tupos <> sArxn then
    raise ESuvtaktikoLa8os.CreateFmt(sPerimevaMetablntesHArxh,
      [p.dk.TrSus.Kwdikas], p.dk.TrSus.Perioxn);
  if p.dk.TrSus.Tupos <> sArxn then
    raise ESuvtaktikoLa8os.CreateFmt(sPerimevaAllnDL,
      [mPKTDL(sArxn), p.dk.TrSus.KwdikasKaiPerigrafn], p.dk.TrSus.Perioxn);
  p.dk.Epomevo;
end;

{ TOvomasmevnSta8era }

function TOvomasmevnSta8era.DesTimn: Variant;
begin
  result := fTimn;
end;

function TOvomasmevnSta8era.Leitourgies: Word;
begin
  result := lsDesTimn or lsSta8ernTimn or lsSta8ernApotimnsn;
end;

constructor TOvomasmevnSta8era.Create(p: TAfnrnmevoProgramma);
{***************************************************************************************************
   TrSus = sEkfrasnAvagvwristikou.
       = _(2^3) + ,    .
***************************************************************************************************}
resourcestring
  sPerimevaIsov = ' =   %s';
var
  ekf: TSustatiko;
begin
  ovoma := p.dk.TrSus;
  p.TY.EgkuroOvoma(ovoma as TEkfrasnAvagvwristikou);
  fArxiko := ovoma;
  if p.dk.Epomevo.Tupos <> sIsov then
    raise ESuvtaktikoLa8os.CreateFmt(sPerimevaIsov, [p.dk.TrSus.Kwdikas], p.dk.TrSus.Perioxn);
  p.dk.Epomevo;                                       // Ekfrasn  TrSus =  ...
  ekf := Ekfrasn(p.dk);                //       TDuadiknPrajn 
  try                          //..    ,      ...
    fTimn := VarAsType(ekf.Timn, ekf.TuposDedomevwv);
  except
    on E: exception do
    begin
      if not (ekf is TAploSustatiko) then                     //     dk,  
        FreeAndNil(ekf);
      if E is ELa8os then
        raise ESuvtaktikoLa8os.Create((E as ELa8os).Message, (E as ELa8os).Perioxn)
      else
        raise ESuvtaktikoLa8os.Create(E.Message, Perioxn);   //  ,   
    end;
  end;
  fTeliko := ekf.Teliko;
  if not (ekf is TAploSustatiko) then
    FreeAndNil(ekf);
end;

function TOvomasmevnSta8era.Perigrafn: WideString;
begin
  case TuposDedomevwv of
    varInt64:  result := ' ';
    varDouble: result := ' ';
    varOleStr: result := ' ';
  else result := ' ';
  end;
end;

function TOvomasmevnSta8era.TuposDedomevwv: TVarType;
begin
  result := VarType(fTimn);
end;

{ TMetablntn }

function TMetablntn.DesTimn: Variant;
resourcestring
  sMnArxikopoinmevnMetablntn = '         '
    + '%s     .';
begin
  if (stoiba[Metatopisn] = nil) or VarIsEmpty(stoiba[Metatopisn]^) then
    raise ELa8osEktelesns.CreateFmt(sMnArxikopoinmevnMetablntn, [ovoma.Kwdikas], Perioxn);
//           
//       TEkfrasnAvagvwristikou.DesTimn     
  result := stoiba[metatopisn]^;
end;

procedure TMetablntn.QeseTimn(const t: Variant);
{***************************************************************************************************
             r <- _(3.14),
  r .      VarType(nTimn)   nTimn    
  ,    UnAssigned.
***************************************************************************************************}
begin
  if VarType(t) = varEmpty then
    stoiba[Metatopisn]^ := Unassigned             //TODO bug:    unassigned 
  else if VarType(t) in [varInt64, varDouble, varOleStr, varBoolean] then
    stoiba[Metatopisn]^ := VarAsType(t, fTuposDedomevwv)
  else
    Assert(true);
end;

function TMetablntn.Leitourgies: Word;
begin
  result := lsDesTimn or lsQeseTimn or lsSta8ernApotimnsn;
end;

constructor TMetablntn.Create(p: TAfnrnmevoProgramma; tupos: TVarType; metatopisn: integer);
{***************************************************************************************************
         .
  TrSus      ,    .
***************************************************************************************************}
begin
  fTuposDedomevwv := tupos;
  fMetatopisn := metatopisn;
  ovoma := p.dk.TrSus;
  stoiba := p.stoiba;
  fArxiko := ovoma;
  fTeliko := ovoma;
  p.dk.Epomevo;
end;

function TMetablntn.Perigrafn: WideString;
begin
  case TuposDedomevwv of
    varInt64:  result := ' ';
    varDouble: result := ' ';
    varOleStr: result := ' ';
  else result := ' ';
  end;
end;

function TMetablntn.TuposDedomevwv: TVarType;
begin
  result := fTuposDedomevwv;
end;

procedure TMetablntn.Arxikopoinse;
{***************************************************************************************************
      .
***************************************************************************************************}
begin
  VarClear(stoiba[Metatopisn]^);
end;

{ TMetablntnPivakas }

procedure TMetablntnPivakas.QeseTimn(const t: Variant);
{***************************************************************************************************
            .
     variant      .
 :
 1.              .
                   
        .
 2.     .         .
***************************************************************************************************}
var
  i: integer;
begin
  Assert(VarIsArray(t) and (VarArrayDimCount(t) = Length(oria) div 2));
  for i := Low(oria) to High(oria) do
    if i mod 2 = 0 then
      Assert(VarArrayLowBound(t, 1 + i div 2) = oria[i])     //   VALB   1
    else
      Assert(VarArrayHighBound(t, 1 + i div 2) = oria[i]);   //   VAHB   1
  stoiba[metatopisn]^ := t;
end;

constructor TMetablntnPivakas.Create(p: TAfnrnmevoProgramma; tupos: TVarType; metatopisn: integer);
{***************************************************************************************************
      ,        [.
     TrSus = ].
***************************************************************************************************}
resourcestring
  sMovoSta8eraOriaPivaka = '      '
         + '      .';
  sMovoAkeraiaOriaPivaka = '      .' + WBR + '%s';
  sMovo8etikaOriaPivaka = '          .'
    + WBR + '%s = %d';
  sPerimevaDejiaAgkuln = '     ';
var
  ekf: TSustatiko;
begin
  fTuposDedomevwv := tupos + varArray;
  fMetatopisn := metatopisn;
  ovoma := p.dk.TrSus;
  stoiba := p.stoiba;
  fArxiko := ovoma;
  fTeliko := ovoma;
  ekf := nil;
  try
    p.dk.Epomevo;       //    [ (       &)
    repeat
      p.dk.Epomevo;                                                      //  [   
      ekf := Ekfrasn(p.dk);
      if not ekf.Ypostnrizei(lsSta8ernTimn) then
        raise ESuvtaktikoLa8os.Create(sMovoSta8eraOriaPivaka, ekf.Perioxn)
      else if ekf.TuposDedomevwv <> varInt64 then
        raise ESuvtaktikoLa8os.CreateFmt(sMovoAkeraiaOriaPivaka, [ekf.KwdikasKaiPerigrafn],
          ekf.Perioxn);
      SetLength(oria, Length(oria) + 2);
      oria[High(oria) - 1] := 1;
      try
        oria[High(oria)] := ekf.Timn; //..    ,     
      except
        on E: exception do
          if E is ELa8os then
            raise ESuvtaktikoLa8os.Create((E as ELa8os).Message, (E as ELa8os).Perioxn)
          else
            raise ESuvtaktikoLa8os.Create(E.Message, ekf.Perioxn);
      end;
      if oria[High(oria)] <= 0 then
        raise ESuvtaktikoLa8os.CreateFmt(sMovo8etikaOriaPivaka, [ekf.Kwdikas, oria[High(oria)]],
         ekf.Perioxn);
    until p.dk.TrSus.Tupos <> sKomma;
    if p.dk.TrSus.Tupos <> sDeAgkuln then
      raise ESuvtaktikoLa8os.Create(sPerimevaDejiaAgkuln, p.dk.TrSus.Perioxn);
    fTeliko := p.dk.TrSus;
    p.dk.Epomevo;                                                        //   
  except
    if not (ekf is TAploSustatiko) then                       //     dk,  
      FreeAndNil(ekf);
    raise;
  end;
end;

function TMetablntnPivakas.Perigrafn: WideString;
begin
  case TuposDedomevwv - varArray of
    varInt64:   result := ' ';
    varDouble:  result := ' ';
    varOleStr:  result := ' ';
    varBoolean: result := ' ';
  else
    Assert(true);
  end;
end;

procedure TMetablntnPivakas.Arxikopoinse;
{***************************************************************************************************
      .
***************************************************************************************************}
begin
  stoiba[metatopisn]^ := VarArrayCreate(oria, VarVariant);                   // 
end;

class function TMetablntnPivakas.SarwseEkfrasn(dk: TDiaxeiristnsKwdika): TSustatiko;
{***************************************************************************************************
   TrSus =         .
***************************************************************************************************}
begin
  if dk.EpSus.Tupos <> sArAgkuln then
    result := inherited SarwseEkfrasn(dk)               // TEkfrasnAvagvwristikou,   
  else
    result := TEkfrasnStoixeiouPivaka.Create(dk);
end;

{ TEkfrasnStoixeiouPivaka }

resourcestring
  sDeiktnsEktosOriwvPivaka =
    ' %d    %s   %d,    .';

function TEkfrasnStoixeiouPivaka.DesTimnHUnassigned: Variant;
resourcestring
  sMnArxikopoinmevosPivakas =  '  %s    .';
var
  i: integer;
  deiktes: array of integer;
begin
  with metablntnPivakas do
  if (stoiba[Metatopisn] = nil) or VarIsEmpty(stoiba[Metatopisn]^) then
    raise ELa8osEktelesns.CreateFmt(sMnArxikopoinmevosPivakas, [ovoma.Kwdikas], Perioxn);
  SetLength(deiktes, orismata.Pln8os);
  for i := 0 to orismata.Pln8os - 1 do with metablntnPivakas do
  begin                                          //   VarArrayLowBound   1
    deiktes[i] := self.orismata[i].Timn;
    if (deiktes[i] < VarArrayLowBound(stoiba[metatopisn]^, i + 1))
     or (deiktes[i] > VarArrayHighBound(stoiba[metatopisn]^, i + 1)) then
     raise ELa8osEktelesns.CreateFmt(sDeiktnsEktosOriwvPivaka, [i + 1, ovoma.Kwdikas, deiktes[i]],
       self.orismata[i].Perioxn);                        //self =>   metablntnPivakas  with
  end;
  result := VarArrayGet(metablntnPivakas.stoiba[metablntnPivakas.Metatopisn]^, deiktes);
end;

function TEkfrasnStoixeiouPivaka.DesTimn: Variant;
resourcestring
  sMnArxikopoinmevoStoixeio = '       %s'
   + '    .';
begin
  result := DesTimnHUnassigned;
  if VarIsEmpty(result) then
   raise ELa8osEktelesns.CreateFmt(sMnArxikopoinmevoStoixeio, [metablntnPivakas.ovoma.Kwdikas],
     Perioxn);
end;

procedure TEkfrasnStoixeiouPivaka.QeseTimn(const t: Variant);
//todo:       ,      ,
//    .       ,      .
var
  i: integer;
  deiktes: array of integer;
begin
  SetLength(deiktes, orismata.Pln8os);
  for i := 0 to orismata.Pln8os - 1 do with metablntnPivakas do
  begin                                          //   VarArrayLowBound   1
    deiktes[i] := self.orismata[i].Timn;
    if (deiktes[i] < VarArrayLowBound(stoiba[metatopisn]^, i + 1))
     or (deiktes[i] > VarArrayHighBound(stoiba[metatopisn]^, i + 1)) then
     raise ELa8osEktelesns.CreateFmt(sDeiktnsEktosOriwvPivaka, [i + 1, ovoma.Kwdikas, deiktes[i]],
       self.orismata[i].Perioxn);                        //self =>   metablntnPivakas  with
  end;
  VarArrayPut(metablntnPivakas.stoiba[metablntnPivakas.metatopisn]^, VarAsType(t, TuposDedomevwv),
   deiktes);
end;

function TEkfrasnStoixeiouPivaka.Leitourgies: Word;
begin
  result := lsDesTimn or lsQeseTimn or lsSta8ernApotimnsn;
end;

constructor TEkfrasnStoixeiouPivaka.Create(dk: TDiaxeiristnsKwdika);
{***************************************************************************************************
   TrSus = .     ""      [.
***************************************************************************************************}
resourcestring
  sOxiAkeraiosDeiktnsPivaka = ' %d    %s   .' + WBR + '%s';
  sPerimevaKiAllovDeiktn = '     %s  %d .';
  sPerimevaDejiaAgkuln = '      .';
var
  i: integer;
begin
  inherited Create;                                              //    
  fArxiko := dk.TrSus;
  metablntnPivakas := (dk.TrSus as TEkfrasnAvagvwristikou).avagvwristiko as TMetablntnPivakas;
  dk.Epomevo;                                                                       //  
  for i := 0 to Length(metablntnPivakas.oria) div 2 - 1 do
  begin
    dk.Epomevo;                                               //      
    orismata.Eisagwgn(Ekfrasn(dk));
    if orismata[i].TuposDedomevwv <> varInt64 then
      raise ESuvtaktikoLa8os.CreateFmt(sOxiAkeraiosDeiktnsPivaka, [orismata.Pln8os,
        metablntnPivakas.ovoma.Kwdikas, orismata[i].KwdikasKaiPerigrafn], orismata[i].Perioxn);
    if (dk.TrSus.Tupos <> sKomma) and (i <> Length(metablntnPivakas.oria) div 2 - 1) then
      raise ESuvtaktikoLa8os.CreateFmt(sPerimevaKiAllovDeiktn,
        [metablntnPivakas.ovoma.Kwdikas, Length(metablntnPivakas.oria) div 2], dk.TrSus.Perioxn);
  end;
  if dk.TrSus.Tupos <> sDeAgkuln then
    raise ESuvtaktikoLa8os.Create(sPerimevaDejiaAgkuln, dk.TrSus.Perioxn);
  fTeliko := dk.TrSus;
  dk.Epomevo;
end;

function TEkfrasnStoixeiouPivaka.Perigrafn: WideString;
begin
  case TuposDedomevwv of
    varInt64:   result := '  ';
    varDouble:  result := '  ';
    varOleStr:  result := '  ';
  else result := '  ';
  end;
end;

function TEkfrasnStoixeiouPivaka.TuposDedomevwv: Word;
begin
  result := metablntnPivakas.TuposDedomevwv - varArray;      //  varArray,   
end;

{ TYpoprogramma }

procedure TYpoprogramma.EisagwgnAvagvwristikou(av: TAvagvwristiko);
{***************************************************************************************************
            av  
   ,      .
***************************************************************************************************}
resourcestring
  sJavadnlwmevoAvagvwristiko = '  %s  .' + WBR + '%s: %s.';
var
  java: TAvagvwristiko;                                       //-,    java ;-)
begin
  java := AvazntnsnAvagvwristikou(av.ovoma.Kwdikas);
  if java <> nil then
    raise ESuvtaktikoLa8os.CreateFmt(sJavadnlwmevoAvagvwristiko,
      [av.ovoma.Kwdikas, java.ovoma.Kwdikas, java.Perigrafn], av.ovoma.Perioxn);
//todo:  kwdikas          ...
  fAvagvwristika.Eisagwgn(av);
end;

function TYpoprogramma.AvazntnsnAvagvwristikou(ovoma: WideString): TAvagvwristiko;
{***************************************************************************************************
             kwdikas.
***************************************************************************************************}
var
  i: integer;
  atovoOvoma: WideString;
begin
  atovoOvoma := KefalaioAtovo(ovoma);
  if (self.ovoma <> nil) and (atovoOvoma = KefalaioAtovo(self.ovoma.Kwdikas)) then
  begin                                              //       
    result := self;
    Exit;
  end;
  for i := 0 to fAvagvwristika.Pln8os - 1 do
    if (atovoOvoma = KefalaioAtovo((fAvagvwristika[i] as TAvagvwristiko).ovoma.Kwdikas)) then
    begin
      result := fAvagvwristika[i] as TAvagvwristiko;
      Exit;
    end;
  result := nil;
end;

procedure TYpoprogramma.Eisagwgn(evtoln: TEvtoln);
{***************************************************************************************************
     .      ,   .
***************************************************************************************************}
var
  i: integer;
begin
  i := Length(evtoles);
  if (i > 0) then                                                      //     
  begin
    if evtoles[i - 1].nEpomevnEvtoln = nil then    //       
      evtoles[i - 1].nEpomevnEvtoln := evtoln;
    if (evtoles[i - 1] is TEvtolnDiakladwsns)
     and ((evtoles[i - 1] as TEvtolnDiakladwsns).nEvtolnDiakladwsns = nil) then
      (evtoles[i - 1] as TEvtolnDiakladwsns).nEvtolnDiakladwsns := evtoln;
  end;
  SetLength(evtoles, i + 1);
  evtoles[i] := evtoln;
  te := evtoln;
end;

procedure TYpoprogramma.W8nsn(evtoln: TEvtoln);
{***************************************************************************************************
     .    , , _,
 , , _, , _.
***************************************************************************************************}
begin
  SetLength(stoibaEmfwleumevwv, Length(stoibaEmfwleumevwv) + 1);
  stoibaEmfwleumevwv[High(stoibaEmfwleumevwv)] := evtoln;
end;

function TYpoprogramma.Ejw8nsn: TEvtoln;
{***************************************************************************************************
      ,      ,
  nil     .
      _.
***************************************************************************************************}
begin
  result := Krufokoitagma;
  SetLength(stoibaEmfwleumevwv, Length(stoibaEmfwleumevwv) - 1);
end;

function TYpoprogramma.Krufokoitagma: TEvtoln;
{***************************************************************************************************
  Peek     .
***************************************************************************************************}
begin
  if Length(stoibaEmfwleumevwv) = 0 then
    result := nil                                           //    SarwseSwma
  else result := stoibaEmfwleumevwv[High(stoibaEmfwleumevwv)];               //  
end;

procedure TYpoprogramma.EgkuroOvoma(ea: TEkfrasnAvagvwristikou);
resourcestring
  sOxiAvagvwristikoMeToOvomaProgrammatos =
    '        .';
  sOxiAvagvwristikoMeToOvomaYpoprogrammatos =
    '        .';
begin
  if ea.avagvwristiko <> nil then                                          //   
    if ea.avagvwristiko.ovoma = p.kp.ovoma then
      raise ESuvtaktikoLa8os.Create(sOxiAvagvwristikoMeToOvomaProgrammatos, ea.Perioxn)
    else if ea.avagvwristiko.ovoma = p.TY.ovoma then
      raise ESuvtaktikoLa8os.Create(sOxiAvagvwristikoMeToOvomaYpoprogrammatos, ea.Perioxn);
end;

procedure TYpoprogramma.AvaluseParametrous;
{***************************************************************************************************
           .
     / .
***************************************************************************************************}
begin
end;

constructor TYpoprogramma.Create(p: TAfnrnmevoProgramma);
begin
  self.p := p;
  fAvagvwristika := TListaSustatikwv.Create;
  p.fTY := self;
  fArxiko := p.dk.TrSus;
  SarwseEvtolnYpoprogramma;
  TEvtolnDnlwseis.Create(p);                                         //    evtoles
  AvaluseParametrous;
end;

destructor TYpoprogramma.Destroy;
var
  i: integer;
begin
  for i := 0 to High(evtoles) do
    FreeAndNil(evtoles[i]);                     //     
  FreeAndNil(fAvagvwristika);
  inherited;
end;

procedure TYpoprogramma.SarwseSwma;
{***************************************************************************************************
     _
***************************************************************************************************}
resourcestring
  sDevEkleiseHEvtoln = '        %s.';
  sAgvwstnEvtoln = '         .' + WBR + '%s';
  sPeriptwsnMetaApoEpileje =
    '          .';
  sMovoYpoprogrammaMetaTelosProgrammatos =
    '            ';
begin
  p.fTY := self;
  p.dk.TrSus := evtoles[High(evtoles)].Teliko;      // evtoles[High(evtoles)] = TEvtolnDnlwseis
  p.dk.Epomevo;                                                        //    ޻
  repeat
    case p.dk.TrSus.Tupos of
      sAvagvwristiko:     TEvtolnAva8esnTimns.Create(p);
      sDiabase:           TEvtolnDiabase.Create(p);
      sGraye:             TEvtolnGraye.Create(p);
      sAv:                TEvtolnAv.Create(p);
      sAlliws_Av:         TEvtolnAlliwsAv.Create(p);
      sAlliws:            TEvtolnAlliws.Create(p);
      sTelos_Av:          TEvtolnTelosAv.Create(p);
      sEpileje:           TEvtolnEpileje.Create(p);
      sPeriptwsn:    if p.dk.EpSus.Tupos <> sAlliws then
                          TEvtolnPeriptwsn.Create(p)
                     else TEvtolnPeriptwsnAlliws.Create(p);
      sTelos_Epilogwv:    TEvtolnTelosEpilogwv.Create(p);
      sOso:               TEvtolnOso.Create(p);
      sGia:               TEvtolnGia.Create(p);
      sTelos_Epavalnyns:  TEvtolnTelosEpavalnyns.Create(p);
      sArxn_Epavalnyns:   TEvtolnArxnEpavalnyns.Create(p);
      sMexris_Otou:       TEvtolnMexrisOtou.Create(p);
      sKalese:            TEvtolnKalese.Create(p);
      sTelos_Programmatos:TEvtolnTelosProgrammatos.Create(p);
      sTelos_Diadikasias: TEvtolnTelosDiadikasias.Create(p);
      sTelos_Suvartnsns:  TEvtolnTelosSuvartnsns.Create(p);
      sTelosArxeiou:      raise ESuvtaktikoLa8os.CreateFmt(sDevEkleiseHEvtoln,
                            [Krufokoitagma.Kwdikas], p.dk.TrSus.Perioxn);
    else
      raise ESuvtaktikoLa8os.CreateFmt(sAgvwstnEvtoln, [p.dk.TrSus.KwdikasKaiPerigrafn],
        p.dk.TrSus.Perioxn);
    end;
    if (te is TEvtolnEpileje) and (p.dk.TrSus.Tupos <> sPeriptwsn) then
      raise ESuvtaktikoLa8os.Create(sPeriptwsnMetaApoEpileje, p.dk.TrSus.Perioxn)
    else if ((te is TEvtolnTelosProgrammatos) or (te is TEvtolnTelosDiadikasias)
     or (te is TEvtolnTelosSuvartnsns))
     and not (p.dk.TrSus.Tupos in [sDiadikasia, sSuvartnsn, sTelosArxeiou]) then
      raise ESuvtaktikoLa8os.Create(sMovoYpoprogrammaMetaTelosProgrammatos, p.dk.TrSus.Perioxn);
  until p.TY.Krufokoitagma = nil;                        //     
  fTeliko := p.dk.TrSus;
end;

procedure TYpoprogramma.Klnsn;
{***************************************************************************************************
   call stack,      1 .Ektelesn.
    call stack            TAfnrnmevoProgramma.
        callstack (    heap),  TYpoprogramma.Klnsn
    Delphi (  ).        
  stack overflow.
***************************************************************************************************}
resourcestring
  sYperxeilnsnStoibas = '       %s.'
    + '      1000  .';
var
  palioYp: TYpoprogramma;
  paliaTE: TEvtoln;
  paliosBP: integer;
  i: integer;
begin
  paliaTE := p.TE;
  palioYp := p.TY;
  if Length(p.callstack) > 1000 then
    raise ELa8osEktelesns.CreateFmt(sYperxeilnsnStoibas, [self.ovoma.Kwdikas], p.TE.Perioxn);
  p.fTY := self;
  SetLength(p.callstack, Length(p.callstack) + 1);
  p.callstack[High(p.callstack)].upoprogramma := self;
  p.callstack[High(p.callstack)].BP := p.stoiba.SP;
  paliosBP := p.stoiba.BP;
  p.stoiba.BP := p.stoiba.SP;
  for i := 0 to fAvagvwristika.Pln8os - 1 do
    if (fAvagvwristika[i].InheritsFrom(TMetablntn))                   //    
     and ((fAvagvwristika[i] as TMetablntn).metatopisn >= 0) then         //   
      p.stoiba.Push(UnAssigned);
//todo: :           ...
  for i := 0 to fAvagvwristika.Pln8os - 1 do
    if (fAvagvwristika[i].InheritsFrom(TMetablntn))                   //    
     and ((fAvagvwristika[i] as TMetablntn).metatopisn >= 0) then         //   
      (fAvagvwristika[i] as TMetablntn).Arxikopoinse;
  te := evtoles[0];
  while te <> nil do                           // ,     loop  !!!
    te := te.Ektelesn;
  for i := fAvagvwristika.Pln8os - 1 downto 0 do
    if (fAvagvwristika[i].InheritsFrom(TMetablntn))                   //    
     and ((fAvagvwristika[i] as TMetablntn).metatopisn >= 0) then         //   
      p.stoiba.Pop;                                                    //    
  p.stoiba.BP := paliosBP;
  SetLength(p.callstack, Length(p.callstack) - 1);
  p.fTY := palioYp;
  if p.TY <> nil then
    p.TY.te := paliaTE;
end;

class function TYpoprogramma.SarwseEkfrasn(dk: TDiaxeiristnsKwdika): TSustatiko;
{***************************************************************************************************
   class function SarwseEkfrasn    TAvagvwristiko   
 .            ,  
            +   
 _  1.   override   TDiadikasia.
***************************************************************************************************}
resourcestring
  sOxiDiadikasiesSeEkfraseis = '         '
    + '   ,          .';
begin
  raise ESuvtaktikoLa8os.Create(sOxiDiadikasiesSeEkfraseis, dk.TrSus.Perioxn);
end;

{ TKuriwsProgramma }

procedure TKuriwsProgramma.SarwseEvtolnYpoprogramma;
begin
  TEvtolnProgramma.Create(p);
end;

function TKuriwsProgramma.Perigrafn: WideString;
begin
  result := mPKTDL(sProgramma);
end;

{ TEkfrasnKuriwsProgrammatos }

constructor TEkfrasnKuriwsProgrammatos.Create(dk: TDiaxeiristnsKwdika);
begin
  inherited Create;
  upoprogramma := (dk.TrSus as TEkfrasnAvagvwristikou).avagvwristiko as TKuriwsProgramma;
  fArxiko := dk.TrSus;
  fTeliko := dk.TrSus;
  dk.Epomevo;
end;

{ TDiadikasia }

procedure TDiadikasia.AvaluseParametrous;
{***************************************************************************************************
        .       
          .
***************************************************************************************************}
resourcestring
  sMnDnlwmevnParametros = '      '
    + '     %s.';
var
  i: integer;
begin
  for i := 0 to High(parametroi) do
  begin
    parametroi[i].avagvwristiko := AvazntnsnAvagvwristikou(parametroi[i].Kwdikas);
    if parametroi[i].avagvwristiko = nil then
      raise ESuvtaktikoLa8os.CreateFmt(sMnDnlwmevnParametros, [parametroi[i].Kwdikas],
        parametroi[i].Perioxn);
  end;
end;

procedure TDiadikasia.SarwseEvtolnYpoprogramma;
begin
  TEvtolnDiadikasia.Create(p);                                       //    evtoles
end;

function TDiadikasia.Perigrafn: WideString;
begin
  result := mPKTDL(sDiadikasia);
end;

{ TEkfrasnDiadikasias }

constructor TEkfrasnDiadikasias.Create(dk: TDiaxeiristnsKwdika);
{***************************************************************************************************
   TrSus =     (  ).    
       '     dk,     
  TDiadikasia.
***************************************************************************************************}
resourcestring
  sPerimevaAristernParev8esn =
    '        %s.';
  sAsumbatnPragmatiknParametros = '     '
    + '      :' + WBR + '%s' + WBR + '%s';
  sLa8osDiastaseisParametrouPivaka = '      -  %s'
    + '        -  %s.';
  sLa8osOriaParametrouPivaka = '     -  %s'
    + '        -  %s.';
  sPerimevaKommaKiAllnParametro = '       .';
  sPerimevaDejiaParev8esn = '  ,   .';
var
  ekf, p: TSustatiko;
  i, j: integer;
  mp: TMetablntnPivakas;
begin
  inherited Create;
  upoprogramma := (dk.TrSus as TEkfrasnAvagvwristikou).avagvwristiko as TDiadikasia;
//  fArxiko := upoprogramma.Arxiko; thank you   for the bug report
  fArxiko := dk.TrSus;
  dk.Epomevo;                                                   //    
  if (Length(upoprogramma.parametroi) = 0) and (dk.TrSus.Tupos <> sArParev8esn) then
  begin                                                //      
    fTeliko := fArxiko;
    Exit;
  end;
  if dk.TrSus.Tupos <> sArParev8esn then
    raise ESuvtaktikoLa8os.CreateFmt(sPerimevaAristernParev8esn, [upoprogramma.ovoma.Kwdikas],
      dk.TrSus.Perioxn);
  for i := 0 to High(upoprogramma.parametroi) do
  begin
    dk.Epomevo;                                        //      
    ekf := Ekfrasn(dk);
    orismata.Eisagwgn(ekf);
// .    if    vartype  ,  
// "base"          .
    if ekf.TuposDedomevwv <> upoprogramma.parametroi[i].TuposDedomevwv then
//         ,  
      if (ekf.TuposDedomevwv = varInt64) and (upoprogramma.parametroi[i].TuposDedomevwv = varDouble)
        and ((self is TEkfrasnSuvartnsns) or not ekf.Ypostnrizei(lsQeseTimn)) then
          // .   or not       (+0)
      else
        raise ESuvtaktikoLa8os.CreateFmt(sAsumbatnPragmatiknParametros,
          [ekf.KwdikasKaiPerigrafn, upoprogramma.parametroi[i].KwdikasKaiPerigrafn], ekf.Perioxn)
    else if ekf.TuposDedomevwv and varArray <> 0 then
    begin
      p := ekf;
      while p is TEkfrasnSeParev8esn do
        p := (ekf as TEkfrasnSeParev8esn).Orisma;
      Assert(p is TEkfrasnAvagvwristikou);
      Assert((p as TEkfrasnAvagvwristikou).avagvwristiko is TMetablntnPivakas);
      mp := (p as TEkfrasnAvagvwristikou).avagvwristiko as TMetablntnPivakas;
      if Length(mp.oria) <>
       Length((upoprogramma.parametroi[i].avagvwristiko as TMetablntnPivakas).oria) then
        raise ESuvtaktikoLa8os.CreateFmt(sLa8osDiastaseisParametrouPivaka,
          [mp.ovoma.Kwdikas, upoprogramma.parametroi[i].Kwdikas], ekf.Perioxn);
      for j := 0 to High(mp.oria) do
        if mp.oria[j] <> (upoprogramma.parametroi[i].avagvwristiko as TMetablntnPivakas).oria[j]
          then raise ESuvtaktikoLa8os.CreateFmt(sLa8osOriaParametrouPivaka,
             [mp.ovoma.Kwdikas, upoprogramma.parametroi[i].Kwdikas], ekf.Perioxn);
    end;
    if (i < High(upoprogramma.parametroi)) and (dk.TrSus.Tupos <> sKomma) then
      raise ESuvtaktikoLa8os.Create(sPerimevaKommaKiAllnParametro, dk.TrSus.Perioxn);
  end;
  if dk.TrSus.Tupos <> sDeParev8esn then
    raise ESuvtaktikoLa8os.Create(sPerimevaDejiaParev8esn, dk.TrSus.Perioxn);
  fTeliko := dk.TrSus;
  dk.Epomevo;
end;

{ TSuvartnsn }

function TSuvartnsn.DesTimn: Variant;
{***************************************************************************************************
     result,            .
      .
***************************************************************************************************}
begin
  result := p.stoiba[-Length(parametroi) - 1]^;
end;

procedure TSuvartnsn.QeseTimn(const t: Variant);
{***************************************************************************************************
     ovomaSuvartnsns := timn     
     .
***************************************************************************************************}
begin
  Assert(VarType(t) in [varInt64, varDouble, varOleStr, varBoolean]);
  p.stoiba[-Length(parametroi) - 1]^ := VarAsType(t, fTuposDedomevwv);
end;

function TSuvartnsn.Leitourgies: Word;
{***************************************************************************************************
    .
***************************************************************************************************}
begin
  result := lsDesTimn or lsQeseTimn or lsSta8ernApotimnsn;
end;

function TSuvartnsn.Perigrafn: WideString;
begin
  case TuposDedomevwv of
    varInt64:  result := ' ';
    varDouble: result := ' ';
    varOleStr: result := ' ';
  else result := ' ';
  end;
end;

function TSuvartnsn.TuposDedomevwv: TVarType;
begin
  result := fTuposDedomevwv;
end;

procedure TSuvartnsn.SarwseEvtolnYpoprogramma;
begin
  TEvtolnSuvartnsn.Create(p);                                        //    evtoles
end;

class function TSuvartnsn.SarwseEkfrasn(dk: TDiaxeiristnsKwdika): TSustatiko;
{***************************************************************************************************
   TrSus =     .       
    '     dk,      
 TSuvartnsn.
***************************************************************************************************}
resourcestring
  sOxiOvomaSuvartnsnsEdw = '        %s   '
    + ' ,     .';
var
  s: TSuvartnsn;
  p: TAfnrnmevoProgramma;
begin
  s := ((dk.TrSus as TEkfrasnAvagvwristikou).avagvwristiko as TSuvartnsn);
  p := s.p;
  if p.seAva8esnTimns then
    if p.TY = s then
    begin
      result := dk.TrSus;                               //   ,   
      dk.Epomevo;                                                //    
    end
    else
      raise ESuvtaktikoLa8os.CreateFmt(sOxiOvomaSuvartnsnsEdw, [dk.TrSus.Kwdikas], dk.TrSus.Perioxn)
  else result := TEkfrasnSuvartnsns.Create(dk);
end;

{ TEkfrasnSuvartnsns }

function TEkfrasnSuvartnsns.DesTimn: Variant;
{***************************************************************************************************
  TEkfrasnSuvartnsns.DesTimn   .
***************************************************************************************************}
resourcestring
  sMnAva8esnTimnsSeSuvartnsn = '       . '
    + '     %s %s ޻.';
var
  i: integer;
begin
  upoprogramma.p.stoiba.Push(UnAssigned);
  for i := 0 to orismata.Pln8os - 1 do
//  UnAssigned      
    if (orismata[i] is TEkfrasnAvagvwristikou)
     and ((orismata[i] as TEkfrasnAvagvwristikou).avagvwristiko is TMetablntn) then
      upoprogramma.p.stoiba.Push(upoprogramma.p.stoiba[
       ((orismata[i] as TEkfrasnAvagvwristikou).avagvwristiko as TMetablntn).metatopisn]^)
    else if (orismata[i] is TEkfrasnStoixeiouPivaka) then 
      upoprogramma.p.stoiba.Push((orismata[i] as TEkfrasnStoixeiouPivaka).DesTimnHUnassigned)
    else upoprogramma.p.stoiba.Push(orismata[i].Timn);
  upoprogramma.Klnsn;
  for i := orismata.Pln8os - 1 downto 0 do      
    upoprogramma.p.stoiba.Pop;           //    ,   
  result := upoprogramma.p.stoiba.Pop;
  if VarIsEmpty(result) then
    raise ELa8osEktelesns.CreateFmt(sMnAva8esnTimnsSeSuvartnsn,
      [upoprogramma.ovoma.Kwdikas, AnsiTelestesSeUnicode('<-')], Perioxn);
  upoprogramma.p.ArxnEvtolns;   //        
end;

function TEkfrasnSuvartnsns.Leitourgies: Word;
{***************************************************************************************************
       lsDesTimn.
***************************************************************************************************}
begin
  result := lsDesTimn;
end;

function TEkfrasnSuvartnsns.TuposDedomevwv: TVarType;
begin
  result := upoprogramma.TuposDedomevwv;
end;

{ TAfnrnmevoProgramma }

procedure TAfnrnmevoProgramma.DiadikasiaAvalusnsAvagvwristikou(eav: TEkfrasnAvagvwristikou);
var
  i: integer;
  atovoOvoma: WideString;
begin
  if TY <> nil then
    eav.avagvwristiko := TY.AvazntnsnAvagvwristikou(eav.Kwdikas)
  else
    eav.avagvwristiko := nil;                 //,     -
  if eav.avagvwristiko = nil then
  begin
    atovoOvoma := KefalaioAtovo(eav.Kwdikas);
    for i := 0 to High(upoprogrammata) do
      if (upoprogrammata[i] <> nil ) and (upoprogrammata[i].ovoma <> nil)        //  
           // ,    SetLength      
       and (atovoOvoma = KefalaioAtovo(upoprogrammata[i].ovoma.Kwdikas)) then
      begin
        eav.avagvwristiko := upoprogrammata[i];
        Exit;
      end;
  end;
end;

procedure TAfnrnmevoProgramma.Mndevise;
var
  i: integer;
begin
  for i := 0 to High(upoprogrammata) do
    FreeAndNil(upoprogrammata[i]);              //     
  upoprogrammata := nil;                                          // SetLength(upoprogrammata, 0)
  callstack := nil;
  stoiba.Arxikopoinse;                 //  ,      
  fKP := nil;
  fTY := nil;
end;

function TAfnrnmevoProgramma.DesTE: TEvtoln;
begin
  if TY <> nil then result := TY.te
  else result := nil;
end;

constructor TAfnrnmevoProgramma.Create;
begin
  dk := TDiaxeiristnsKwdika.Create(DiadikasiaAvalusnsAvagvwristikou);
  fStoiba := TStoiba.Create;
end;

destructor TAfnrnmevoProgramma.Destroy;
begin
  Mndevise;
  FreeAndNil(fStoiba);
  FreeAndNil(dk);
  inherited;
end;

procedure TAfnrnmevoProgramma.Ektelesn;
begin
  callstack := nil;
  stoiba.Arxikopoinse;                 //  ,      
  fTY := nil;                                           //    upoprogrammata[0].Klnsn
  upoprogrammata[0].Klnsn;
end;

procedure TAfnrnmevoProgramma.Metaglwttisn;
{***************************************************************************************************
    dk    .
  '        ,    
 upoprogramma[0].Create.
     dk.EpomevoYpoprogramma      .
        SarwseSwma,     .
***************************************************************************************************}
resourcestring
  sYpoprogrammataMetaKuriwsProgramma =
    '        .';
var
  i: integer;
begin
  while dk.Epomevo.Tupos = sTelosGrammns do                //      
    ;                                                  //      
  if dk.TrSus.Tupos <> sProgramma then
    if dk.TrSus.Tupos in [sDiadikasia, sSuvartnsn] then
      raise ESuvtaktikoLa8os.Create(sYpoprogrammataMetaKuriwsProgramma, dk.TrSus.Perioxn)
    else
    raise ESuvtaktikoLa8os.CreateFmt(sPerimevaAllnDL,
      [mPKTDL(sProgramma), dk.TrSus.KwdikasKaiPerigrafn], dk.TrSus.Perioxn);  
  i := 0;
  SetLength(upoprogrammata, i + 1);
  upoprogrammata[i] := TKuriwsProgramma.Create(self);
  while dk.EpomevoYpoprogramma do
  begin
    Inc(i);
    SetLength(upoprogrammata, i + 1);
    if dk.TrSus.Tupos = sDiadikasia then
      upoprogrammata[i] := TDiadikasia.Create(self)
    else
      upoprogrammata[i] := TSuvartnsn.Create(self);
  end;
  for i := 0 to High(upoprogrammata) do
    upoprogrammata[i].SarwseSwma;              //       
  fTY := KP;
  TY.te := nil;
end;

end.
