{***************************************************************************************************
 This file is part of the GLOSSA Programming Language Interpreter (  )
 Copyright (C) 2002-2008 Alkis Georgopoulos - http://users.sch.gr/alkisg/

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************************************}
unit uLa8os;

interface

uses SysUtils, Windows{$IFNDEF CONSOLE_GLOSSA}, TntSysUtils{$ENDIF};

type
  TSnmeio = type TPoint;

  TPerioxn = packed record
    case Integer of
      0: (X1, Y1, X2, Y2: Longint);
      1: (Arxn, Telos: TSnmeio);
  end;

{$IFDEF CONSOLE_GLOSSA}
  WideException = class(Exception);
{$ENDIF}
  ELa8os = class(WideException)
  protected
    fPerioxn: TPerioxn;
  public
    property Perioxn: TPerioxn read fPerioxn;
    constructor Create; overload;
    constructor Create(const Msg: WideString); overload;
    constructor Create(const Msg: WideString; const perioxn: TPerioxn); overload;
    constructor CreateFmt(const Msg: WideString; const Args: array of const;
      const perioxn: TPerioxn); overload;
  end;

  ELektikoLa8os = class(ELa8os)
  end;

  ESuvtaktikoLa8os = class(ELa8os)
  end;

  ELa8osEktelesns = class(ELa8os)
  end;

  function ErrorAddress(address: Pointer): WideString;

const
  Snmeio0: TSnmeio = (X: 0; Y: 0);
  Pou8eva: TPerioxn = (X1: 0; Y1: 0; X2: 0; Y2: 0);      //     

function KavePerioxn(const X1, Y1, X2, Y2: Integer): TPerioxn;

{**************************************************************************************************}
implementation
{**************************************************************************************************}

uses uLektikosAvalutns;

function KavePerioxn(const X1, Y1, X2, Y2: Integer): TPerioxn; inline;
begin
  Result.X1 := X1;
  Result.Y1 := Y1;
  Result.X2 := X2;
  Result.Y2 := Y2;
end;

constructor ELa8os.Create;
begin
  fPerioxn := Pou8eva;
  inherited Create('');
end;

constructor ELa8os.Create(const Msg: WideString);
begin
  fPerioxn := Pou8eva;
  inherited Create(Msg);
end;

constructor ELa8os.Create(const Msg: WideString; const perioxn: TPerioxn);
begin
  fPerioxn := perioxn;
  inherited Create(Msg);
end;

constructor ELa8os.CreateFmt(const Msg: WideString; const Args: array of const;
  const perioxn: TPerioxn);
begin
  fPerioxn := perioxn;
  inherited CreateFmt(Msg, Args);
end;

function ErrorAddress(address: Pointer): WideString;
begin
{$IFDEF Release}
  Result := '';
{$ELSE}
  Result := WBR + ': ' + IntToHex(Integer(address), 8);
{$ENDIF}
end;

initialization
end.
