{***************************************************************************************************
 This file is part of the GLOSSA Programming Language Interpreter (  )
 Copyright (C) 2002-2008 Alkis Georgopoulos - http://users.sch.gr/alkisg/

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************************************}
unit uProgramma;

interface

uses uLa8os, uSustatika, uEvtoles, SysUtils, Variants;

type
  IDiepafnProgrammatos = interface(IInterface)
// reference counting,        memory leaks,
//  ,       TInterfacedObject
    procedure DiadikasiaEjodou(const keimevo: WideString);
    function  DiadikasiaEisodou(const mnvuma: WideString): WideString;
    procedure DiadikasiaArxnsEvtolns;
  end;

{***************************************************************************************************
  programma.katastasn      . Read-only.
  programma.aitnsn         . (Read) - Write.
 :
  *    aiTermatismos      katastasn = kpTermatismevo.
  *  step in  aiBnmaMeEisodoSeYpoprogrammata   ArxnEvtolns    kpPausn.
  *  step in  aiBnmaMeYperbasnYpoprogrammatwv +   callstack   ArxnEvtolns 
      kpPausn.
  *  step out  aiEktelesnWsEpistrofnYpoprogrammatos +   callstack  
    ArxnEvtolns    kpPausn.
***************************************************************************************************}
  TKatastasnProgrammatos = (kpAmetaglwttisto, kpMetaglwttismevo, kpTermatismevo = kpMetaglwttismevo,
   kpEktelesn);
  TAitnsnKatastasns = (aiKamia, aiAmetaglwttisto, aiSuvexisnEktelesns, aiPausnEktelesns,
  aiBnmaMeEisodoSeYpoprogrammata, aiBnmaMeYperbasnYpoprogrammatwv,
  aiEktelesnWsEpistrofnYpoprogrammatos, aiTermatismosEktelesns, aiEktelesnProteivomevnsLusns);

  TProgramma = class(TAfnrnmevoProgramma)
{***************************************************************************************************
  TProgramma   TAfnrnmevoProgramma        
 .
        frmGlwssa,    , ,
  /,   , /   .
           uMain  3-4 callbacks   .
***************************************************************************************************}
  protected
    fKatastasn: TKatastasnProgrammatos;
    fAitnsn: TAitnsnKatastasns;
    ba8osAitnsns: integer;         //   call stack    
    fArEktEvtolwv: Int64;
    iDiepafn: IDiepafnProgrammatos;
    procedure   QeseAitnsn(const aitnsn: TAitnsnKatastasns);
    procedure   Ejodos(const keimevo: WideString); override;
    procedure   Eisodos(ekfMet: TSustatiko); override;
    procedure   ArxnEvtolns; override;
    procedure   TermatismosEktelesns;      //       ,  raise
  public
    la8os: record
      perioxn: TPerioxn;
      tupos: WideString;
      mnvuma: WideString;
    end;
    titlos, ovomaArxeiou: WideString;
    ekfwvnsn, lusn, paradeigmaEisodou, proteivomevnLusn: WideString;
//       .htm  (= HTML),     
//         .
    tuposDiabasmatos: TVarType;                           //   sxeO8ovnXrnstnKeyPress
    seDiabasma: boolean;      //   sxeO8ovnXrnstnKeyPress,    TDiadikasiaEisodou
    sePausn: boolean;
    property    DiaKw: TDiaxeiristnsKwdika read dk;
    property    Katastasn: TKatastasnProgrammatos read fKatastasn;
    property    Aitnsn: TAitnsnKatastasns read fAitnsn write QeseAitnsn;
    property    Kln8evtaYpoprogrammata: TKln8evtaYpoprogrammata read callstack;
    property    ArEktEvtolwv: Int64 read fArEktEvtolwv;
    constructor Create(iDiepafn: IDiepafnProgrammatos);
    destructor  Destroy; override;
    procedure   Arxikopoinsn;
    function    Metaglwttisn(const kwdikas: WideString): boolean;
    procedure   Ektelesn; override;
    procedure   QeseYpoprogrammaParakolou8nsns(deiktns: integer);
  end;

function VariantApoKeimevo(k: WideString; tuposDedomevwv: TVarType): Variant;
procedure TimnSustatikouApoKeimevo(s: TSustatiko; k: WideString);

implementation

uses uLektikosAvalutns, uRu8miseis, Registry, Windows;

procedure TProgramma.Ejodos(const keimevo: WideString);
begin
  iDiepafn.DiadikasiaEjodou(keimevo);
end;

procedure TProgramma.TermatismosEktelesns;
{***************************************************************************************************
          .
***************************************************************************************************}
resourcestring
  sTermatismosLogwAllagwv = '        .';
  sTermatismosApoXrnstn = '       .';
begin
  if fAitnsn = aiAmetaglwttisto then
  begin
    fAitnsn := aiKamia;
    fKatastasn := kpAmetaglwttisto;
    raise ELa8osEktelesns.Create(sTermatismosLogwAllagwv, Pou8eva);
  end
  else if fAitnsn = aiTermatismosEktelesns then
  begin
    fAitnsn := aiKamia;
    fKatastasn := kpTermatismevo;
    raise ELa8osEktelesns.Create(sTermatismosApoXrnstn, Pou8eva);
  end
  else if fKatastasn in [kpAmetaglwttisto, kpTermatismevo] then
    raise ELa8osEktelesns.Create                                 //   
  else
    Assert(true);
end;

procedure TProgramma.Eisodos(ekfMet: TSustatiko);
var
  keimevo: WideString;
begin
  if ekfMet <> nil then
  begin
    tuposDiabasmatos := ekfMet.TuposDedomevwv;
    keimevo := '  ' + ekfMet.KwdikasKaiPerigrafn;
  end
  else begin
    tuposDiabasmatos := varEmpty;                                          //   Enter
    keimevo := '    [Enter]...';
  end;
  seDiabasma := true;
  keimevo := iDiepafn.DiadikasiaEisodou(keimevo);
  if (fAitnsn in [aiAmetaglwttisto, aiTermatismosEktelesns])
   or (fKatastasn in [kpAmetaglwttisto, kpTermatismevo]) then
    TermatismosEktelesns;
  if ekfMet <> nil then                              //       Enter
    TimnSustatikouApoKeimevo(ekfMet, keimevo);
end;

constructor TProgramma.Create(iDiepafn: IDiepafnProgrammatos);
begin
  inherited Create;
  self.iDiepafn := iDiepafn;
  Arxikopoinsn;
end;

destructor TProgramma.Destroy;
begin
  if fKatastasn <> kpAmetaglwttisto then
    Mndevise;
  inherited;
end;

procedure TProgramma.Arxikopoinsn;
begin
  ovomaArxeiou := '';
  titlos := ' ';
  ekfwvnsn := '';
  lusn := '';
  paradeigmaEisodou := '';
  proteivomevnLusn := '';
  ovoma := '';
  Mndevise;
  fAitnsn := aiKamia;
  fKatastasn := kpAmetaglwttisto;
  seDiabasma := false;
end;

function TProgramma.Metaglwttisn(const kwdikas: WideString): boolean;
{***************************************************************************************************
          .
***************************************************************************************************}
begin
  Aitnsn := aiAmetaglwttisto;
  try
    dk.Kwdikas := kwdikas;                              //     
    inherited Metaglwttisn;
    fKatastasn := kpTermatismevo;
    la8os.tupos := '   ';
    la8os.mnvuma := '';
    la8os.perioxn := Pou8eva;
  except
    on E: ELektikoLa8os do
    begin
      la8os.tupos := '    ';
      la8os.mnvuma := E.Message + ErrorAddress(ExceptAddr);
      la8os.perioxn := E.perioxn;
    end;
    on E: ELa8os do                          //  ""  ,  
    begin
      la8os.tupos := '    ';
      la8os.mnvuma := E.Message + ErrorAddress(ExceptAddr);
      la8os.perioxn := E.Perioxn;
    end;
    on E: EAssertionFailed do
    begin
      la8os.tupos := ' !';
      la8os.mnvuma := ' :' + WBR
        + '  ' + E.Message + WBR
        + '    email  Alkis.Georgopoulos@gmail.com'
        + '             '
        + '    (..      ).'
        + ErrorAddress(ExceptAddr);
      la8os.perioxn := Pou8eva;
    end;
    on E: Exception do
    begin
      la8os.tupos := '   ';
      la8os.mnvuma := E.Message + ErrorAddress(ExceptAddr);
      la8os.perioxn := Pou8eva;
    end;
  end;
  result := fKatastasn = kpMetaglwttismevo;
  if not result then
    Mndevise;                         //    ( )
end;

procedure TProgramma.Ektelesn;
{***************************************************************************************************
      .         
    .
      ,    AMETAGLWTTISTO,    nil.
***************************************************************************************************}
begin
  Assert(fKatastasn = kpMetaglwttismevo);
  try
    fKatastasn := kpEktelesn;
    fArEktEvtolwv := 0;
    inherited Ektelesn;
    la8os.tupos := '     ';
    la8os.mnvuma := '';
    la8os.perioxn := Pou8eva;
  except
    on E: ELa8osEktelesns do
    begin
      if fKatastasn in [kpAmetaglwttisto, kpTermatismevo] then
      begin
        la8os.tupos := E.Message;                               //     
        la8os.mnvuma := '';
      end
      else begin
        la8os.tupos := '     ';
        la8os.mnvuma := E.Message + ErrorAddress(ExceptAddr);
      end;
      la8os.perioxn := E.Perioxn;
    end;
    on E: ELa8os do
    begin
      la8os.tupos := '     ';
      la8os.mnvuma := E.Message + ErrorAddress(ExceptAddr);
      la8os.perioxn := E.Perioxn;
    end;
    on E: EAssertionFailed do
    begin
      la8os.tupos := ' !';
      la8os.mnvuma := ' :' + WBR
        + '  ' + E.Message + WBR
        + '    email  Alkis.Georgopoulos@gmail.com'
        + '             '
        +  '   (..      ).' + ErrorAddress(ExceptAddr);
      la8os.perioxn := Pou8eva;
    end;
    on E: Exception do
    begin
      la8os.tupos := '     ';
      la8os.mnvuma := E.Message + ErrorAddress(ExceptAddr);
      la8os.perioxn := Pou8eva;
    end;
  end;
//        
  if fKatastasn = kpEktelesn then           //     ( )
    fKatastasn := kpTermatismevo;
  fAitnsn := aiKamia;
  if fKatastasn = kpAmetaglwttisto then               //       objects
    Mndevise;
end;

procedure TProgramma.ArxnEvtolns;
begin
  Inc(fArEktEvtolwv);
  Assert(fKatastasn = kpEktelesn);
  case fAitnsn of
    aiAmetaglwttisto, aiTermatismosEktelesns:
      TermatismosEktelesns;
    aiBnmaMeYperbasnYpoprogrammatwv:
      if Length(callstack) <= ba8osAitnsns then
      begin
        sePausn := true;
        fAitnsn := aiKamia;
      end;
    aiEktelesnWsEpistrofnYpoprogrammatos:
      if Length(callstack) < ba8osAitnsns then
      begin
        sePausn := true;
        fAitnsn := aiKamia;
      end;
  end;
  iDiepafn.DiadikasiaArxnsEvtolns;
  if (fAitnsn in [aiAmetaglwttisto, aiTermatismosEktelesns])          //    ,
   or (fKatastasn in [kpAmetaglwttisto, kpTermatismevo]) then     //    ""
    TermatismosEktelesns                                               // ,  
  else if fAitnsn = aiBnmaMeEisodoSeYpoprogrammata then
  begin
    sePausn := true;                   //     DiadikasiaArxnsEvtolns
    fAitnsn := aiKamia;
  end;
end;

procedure TProgramma.QeseAitnsn(const aitnsn: TAitnsnKatastasns);
begin
  fAitnsn := aitnsn;
  case fAitnsn of
    aiAmetaglwttisto, aiTermatismosEktelesns:
    begin
      seDiabasma := false;                     //    [] ,  
      sePausn := false;                                  //   ArxnEvtolns   Eisodos.
      if (fAitnsn = aiAmetaglwttisto) and (fKatastasn = kpTermatismevo) then
      begin                                       //   ,    sxeLusnChange
        fKatastasn := kpAmetaglwttisto;
        Mndevise;
      end;
      if fKatastasn <> kpEktelesn then
        fAitnsn := aiKamia;
    end;
    aiSuvexisnEktelesns, aiEktelesnProteivomevnsLusns:
    begin
      fAitnsn := aiKamia;                                                     // 
      Assert(fKatastasn <> kpAmetaglwttisto);
      sePausn := false;
      if fKatastasn <> kpEktelesn then
        Ektelesn;
    end;
    aiBnmaMeEisodoSeYpoprogrammata, aiBnmaMeYperbasnYpoprogrammatwv,
    aiEktelesnWsEpistrofnYpoprogrammatos:
    begin
      Assert(fKatastasn <> kpAmetaglwttisto);
      ba8osAitnsns := Length(callstack);
      sePausn := false;
      if fKatastasn <> kpEktelesn then
        Ektelesn;
    end;
    aiPausnEktelesns:
    begin
      fAitnsn := aiKamia;                                                      // 
      Assert(fKatastasn = kpEktelesn);
      sePausn := true;
    end;
  end;
end;

procedure TProgramma.QeseYpoprogrammaParakolou8nsns(deiktns: integer);
{***************************************************************************************************
           
 .  deiktns  0 .. High(callstack)    TY   BP,  deiktns 
    High(callstack).
***************************************************************************************************}
begin
  if (deiktns >= Low(callstack)) and (deiktns <= High(callstack)) then
  begin
    fTY := callstack[deiktns].upoprogramma;
    stoiba.BP := callstack[deiktns].BP;
  end
  else if Length(callstack) = 0 then
  begin
    fTY := KP;                                             // nil     
    stoiba.BP := 0;
  end
  else begin                                              //     
    fTY := callstack[High(callstack)].upoprogramma;
    stoiba.BP := callstack[High(callstack)].BP;
  end
end;

function VariantApoKeimevo(k: WideString; tuposDedomevwv: TVarType): Variant;
resourcestring
  sAduvatnAva8esn = '    .';
var
  i: int64;
  d: double;
  code: integer;
begin
  code := 0;
  if tuposDedomevwv = varInt64 then
  begin
{$IFOPT R+}{$DEFINE RestoreRangeChecking}{$ELSE}{$UNDEF RestoreRangeChecking}{$ENDIF}{$R-}
    Val(k, i, code);
{$IFDEF RestoreRangeChecking}{$R+}{$ENDIF}
    result := VarAsType(i, varInt64);
  end else if tuposDedomevwv = varDouble then
  begin
{$IFOPT R+}{$DEFINE RestoreRangeChecking}{$ELSE}{$UNDEF RestoreRangeChecking}{$ENDIF}{$R-}
    Val(k, d, code);
{$IFDEF RestoreRangeChecking}{$R+}{$ENDIF}
    result := VarAsType(d, varDouble);
  end else
    result := k;
  if code <> 0 then
    raise ESuvtaktikoLa8os.Create(sAduvatnAva8esn, Pou8eva);
end;

procedure TimnSustatikouApoKeimevo(s: TSustatiko; k: WideString);
resourcestring
  sAduvatnAva8esnTimns = '       %s  %s.';
begin
  try
    s.Timn := VariantApoKeimevo(k, s.TuposDedomevwv);
  except
    raise ELa8osEktelesns.CreateFmt(sAduvatnAva8esnTimns, [k, s.KwdikasKaiPerigrafn], s.Perioxn);
  end;
end;

initialization
end.
