{***************************************************************************************************
 This file is part of the GLOSSA Programming Language Interpreter (  )
 Copyright (C) 2002-2008 Alkis Georgopoulos - http://users.sch.gr/alkisg/

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************************************}
unit uSustatika;
{***************************************************************************************************
          :
 TSustatiko, TAploSustatiko, TSuv8etoSustatiko, TAvagvwristiko, TSta8era    "",
 TEvtoln, TDiaxeiristnsKwdika, TYpoprogramma.
       :
 uSarwtnsEkfrasewv: TMovadiaiaPrajn, TEtoimnSuvartnsn, TDiadiknPrajn .
 uProgramma:   .
***************************************************************************************************}
interface

uses
  Windows, Variants, SysUtils, WideStrings, Classes,
  uLektikosAvalutns, uLa8os, uRu8miseis, uDiafora;

const
// .       .
  lsKamia            = $0000;                     //    Kwdika  Perioxn
  lsDesTimn          = $0001;               //  ,       
  lsQeseTimn         = $0002;
  lsSta8ernTimn      = $0004;       //       
  lsSta8ernApotimnsn = $0008; //     ,    

type
  TAploSustatiko = class;                                           //   TSustatiko

  TSustatiko = class
{***************************************************************************************************
  TSustatiko     .
  abstract   Create.        ,
     TAploSustatiko  TSuv8etoSustatiko.
    TEkfrasn,       ('')   (1 + 2.1).
 (       ,     )
***************************************************************************************************}
  protected
    function    DesPerioxn: TPerioxn; virtual; abstract;
    function    Leitourgies: Word; virtual;                                           //: lsKamia
    function    DesTimn: Variant; virtual;                                         //: UnAssigned
    procedure   QeseTimn(const t: Variant); virtual;                      //: La8osProgrammatistn
  public
    function    Perigrafn: WideString; virtual;               //.. DL, SB, SX,   
    function    KwdikasKaiPerigrafn: WideString;           //    
    function    Arxiko: TAploSustatiko; virtual; abstract;                  //  
    function    Teliko: TAploSustatiko; virtual; abstract;             // -  
    function    Kwdikas: WideString; virtual; abstract;
    property    Perioxn: TPerioxn read DesPerioxn;     //    Perioxn.Arxn 
    property    Timn: Variant read DesTimn write QeseTimn;
    function    TuposDedomevwv: TVarType; virtual; //Int64, Double, OleStr, Boolean, [+ array] Empty
    function    Ypostnrizei(ti: Word): boolean;                         //if Leitourgies and ti = ti
  end;

  TAploSustatiko = class(TSustatiko)                          //     
  private                             //  : ,    
    fTupos: TTuposSustatikou;
    fEpomevo: TAploSustatiko;                   //  (   )  
  protected                                  //        dk.
    fKwdikas: WideString;                      //  destroy   ava8esnKwdika  .
    function    DesPerioxn: TPerioxn; override;                        //    nil.
  public
    function    Perigrafn: WideString; override;  //      TEkfrasnAvagvwristikou
    property    Tupos: TTuposSustatikou read fTupos;
    constructor Create(tupos: TTuposSustatikou; kwdikas: WideString);
    function    Arxiko: TAploSustatiko; override;
    function    Teliko: TAploSustatiko; override;
    function    Kwdikas: WideString; override;
  end;

  TSuv8etoSustatiko = class(TSustatiko)                     //     
{***************************************************************************************************
       ,     
    TProgramma.          
  .
***************************************************************************************************}
  protected
    fArxiko, fTeliko: TAploSustatiko;                             //   ,  
    function    DesPerioxn: TPerioxn; override;
  public
    function    Arxiko: TAploSustatiko; override;
    function    Teliko: TAploSustatiko; override;
    function    Kwdikas: WideString; override;
  end;

  TDiaxeiristnsKwdika = class;

  TTelosGrammns = class(TAploSustatiko)
//       .         
//      .
//     TGrammes.      indexof.
//            ,   
//  ,          
//  .
  protected
    dk: TDiaxeiristnsKwdika;
  public
    constructor Create(tupos: TTuposSustatikou; kwdikas: WideString; dk: TDiaxeiristnsKwdika);
  end;

  TSta8era = class(TAploSustatiko)                               //  (1, 1.1, '', )
  protected
    fTimn: Variant;                     //   fTupos,   varType  fTimn
    function    DesTimn: Variant; override;
    function    Leitourgies: Word; override;                                             //lsDesTimn
  public
    function    TuposDedomevwv: TVarType; override;
    procedure   Apotimnse;            //        ,
  end;                                                             //    .

  TAvagvwristiko = class;                              //   TEkfrasnAvagvwristikou

  TEkfrasnAvagvwristikou = class(TAploSustatiko)
{***************************************************************************************************
  TEkfrasnAvagvwristikou     ,   
  TAvagvwristiko.
    ,            
        ,   ..  "".
***************************************************************************************************}
  protected
    function    DesTimn: Variant; override;
    procedure   QeseTimn(const t: Variant); override;
    function    Leitourgies: Word; override;
  public
    avagvwristiko: TAvagvwristiko;            //   TDiadikasiaAvalusnsAvagvwristikou
    function    TuposDedomevwv: TVarType; override;
    function    Perigrafn: WideString; override;                      //  avagvwristiko
  end;

  TAvagvwristiko = class(TSuv8etoSustatiko)
{***************************************************************************************************
   TAvagvwristiko          .
             .
   subclassed     (,  ),    
     :
   *  ,   
   * ,   
   *   ,      
   *      ,    / 
        ,      .
     ,   Create      dk   
  .
   SarwseEkfrasn     Ekfrasn,       
     .           
    (..  (),       
  TEkfrasnAvagvwristikou)     (.. [3, 4] <- 12,    
      TEkfrasnStoixeiouPivaka).
***************************************************************************************************}
  public
    ovoma: TAploSustatiko;                                      //     
    class function SarwseEkfrasn(dk: TDiaxeiristnsKwdika): TSustatiko; virtual;      //:  TrSus
  end;

  TListaSustatikwv = class
{***************************************************************************************************
   .     ,  ,    
      ,         
 .      subclassed    ,  TEvtoln
  TYpoprogramma.  
***************************************************************************************************}
  protected
    sustatika: array of TSustatiko;
    function    DesStoixeio(i: integer): TSustatiko;
    procedure   QeseStoixeio(i: integer; s: TSustatiko);
    function    DesPln8os: integer;
  public
    property    Stoixeia[i: integer]: TSustatiko read DesStoixeio write QeseStoixeio; default;
    property    Pln8os: integer read DesPln8os;
    procedure   Eisagwgn(s: TSustatiko);           //    fArxiko, fTeliko
    destructor  Destroy; override;                       //   ,    dk
  end;

  TDiadikasiaAvalusnsAvagvwristikou = procedure(eav: TEkfrasnAvagvwristikou) of object;
//          TrSus.
//      eav.avagvwristiko

  TDiaxeiristnsKwdika = class
{***************************************************************************************************
              .
           ,     , 
  pointer.       pointers  ,  
       (TList.IndexOf(TSustatiko.Grammn)).  
        pointers ,    "deleted".
   pointer            
  .    .
  TDiaxeiristnsKwdika      TList  ,       
  TList  TSustatiko.
  TSustatiko      ,        
  . ..             
    (  ),    ,    
              TProgramma  
 .       TEvtolnMetablntes      TProgramma,
     full rescan.
          (?).       
 dk     .
***************************************************************************************************}
  private
    fTrSus, fEpSus: TAploSustatiko;             // ,  .  .
    la8osAmp: boolean;                          //     &, 
    telnGrammns: array of TTelosGrammns;               // 2 sTelosArxeiou     
    basnAplwv: TTelosGrammns;        //    ,   
    daa: TDiadikasiaAvalusnsAvagvwristikou;
    procedure   Apeleu8erwse;      //  ,   QeseKwdika    Destroy
    function    EpomevoSnmavtiko: TAploSustatiko;
    procedure   QeseTrSus(const trSus: TAploSustatiko);  // ""    
    function    DesKwdika: WideString;              //      
    procedure   QeseKwdika(const k: WideString);            // la    fGrammes
    procedure   FtiajeKeva(ap: TAploSustatiko; keva: integer);      //  #   ap
  public
    property    TrSus: TAploSustatiko read fTrSus write QeseTrSus;
    property    EpSus: TAploSustatiko read fEpSus;
    property    Kwdikas: WideString read DesKwdika write QeseKwdika;
    constructor Create(daa: TDiadikasiaAvalusnsAvagvwristikou);
    function    Epomevo: TAploSustatiko;     //       &  
    function    EpomevoYpoprogramma: boolean;         //true      
    function    BresGrammn(s: TSustatiko): integer;                                   // s.Arxiko
    function    SAS(s: TSnmeio): TAploSustatiko;                         //(Aplo) SustatikoApoSnmeio
    procedure   MorfopoinseKwdika;
    destructor  Destroy; override;                  //      
  end;

resourcestring
  sAgvwstoAvagvwristiko = ' : %s';

{**************************************************************************************************}
implementation
{**************************************************************************************************}

{ TSustatiko }

function TSustatiko.DesTimn: Variant;
//    exception,      
begin
  result := UnAssigned;
end;

procedure TSustatiko.QeseTimn(const t: Variant);
begin
  Assert(true);
end;

function TSustatiko.Ypostnrizei(ti: Word): boolean;
begin
  result := Leitourgies and ti = ti;
end;

function TSustatiko.Perigrafn: WideString;
begin
  result := ' ';
end;

function TSustatiko.KwdikasKaiPerigrafn: WideString;
begin
  result := '' + Kwdikas + ': ' + Perigrafn;
end;

function TSustatiko.TuposDedomevwv: TVarType;
//    exception,      
begin
  result := varEmpty;
end;

function TSustatiko.Leitourgies: Word;
//            .
begin
  result := lsKamia;
end;

{ TAploSustatiko }

function TAploSustatiko.DesPerioxn: TPerioxn;
{***************************************************************************************************
            s.
     zero-based (   ),     +1 
 '  .
***************************************************************************************************}
var
  a: TAploSustatiko;
begin
  if fTupos = sTelosArxeiou then                          //     ,
  begin                                      //       
    a := (self as TTelosGrammns).dk.telnGrammns[High((self as TTelosGrammns).dk.telnGrammns) - 1];
    if a = self then                                            //     
      Result := Pou8eva
    else begin
      while a.fEpomevo <> self do            //       
        a := a.fEpomevo;
      if (a is TTelosGrammns) then
        Result := KavePerioxn(0, High((self as TTelosGrammns).dk.telnGrammns),
          0, High((self as TTelosGrammns).dk.telnGrammns))
      else
        Result := KavePerioxn(a.Perioxn.X2, a.Perioxn.Y2, a.Perioxn.X2, a.Perioxn.Y2);
    end;
    Exit;
  end;
  a := self;
  while not (a is TTelosGrammns) do                        //     
    a := a.fEpomevo;                                                     //    dk
  result.Y1 := (a as TTelosGrammns).dk.BresGrammn(a as TTelosGrammns);
  result.Y2 := result.Y1;
  a := (a as TTelosGrammns).dk.telnGrammns[result.Y1].fEpomevo;      //  -> 
  result.X1 := 0;
  while a <> self do
  begin
    Inc(result.X1, Length(a.fKwdikas));
    a := a.fEpomevo;
  end;
  result.X2 := result.X1 + Length(self.fKwdikas);
end;

function TAploSustatiko.Perigrafn: WideString;
{***************************************************************************************************
       ,    TEkfrasnAvagvwristikou.
***************************************************************************************************}
const
  pt: array[sTelosArxeiou .. sAvoixtnAlfari8mntiknSt] of WideString =              //Perigrafn Tupou
  (' ', ' ', ' ', ' ', '',
   ' ', ' ', ' ', ' ');
begin
  if (fTupos >= sTelosArxeiou) and (fTupos <= sAvagvwristiko) then result := pt[fTupos]
  else if (fTupos in [sAln8ns, sCeudns]) then result := ' '
  else if (fTupos >= sTRMLow) and (fTupos <= sTRMHigh) then result := ''
  else if (fTupos >=  sDLLow) and (fTupos <=  sDLHigh) then result := ' '
  else Assert(true);
end;

function TAploSustatiko.Arxiko: TAploSustatiko;
begin
  result := self;
end;

function TAploSustatiko.Teliko: TAploSustatiko;
begin
  result := self;
end;

function TAploSustatiko.Kwdikas: WideString;
begin
  result := fKwdikas;
end;

constructor TAploSustatiko.Create(tupos: TTuposSustatikou; kwdikas: WideString);
begin
  fTupos := tupos;
  fKwdikas := kwdikas;
end;

constructor TTelosGrammns.Create(tupos: TTuposSustatikou; kwdikas: WideString;
 dk: TDiaxeiristnsKwdika);
begin
  self.dk := dk;
  inherited Create(tupos, kwdikas);
end;

{ TSuv8etoSustatiko }

function TSuv8etoSustatiko.DesPerioxn: TPerioxn;
begin
  result.Arxn := Arxiko.Perioxn.Arxn;
  result.Telos := Teliko.Perioxn.Telos;
end;

function TSuv8etoSustatiko.Arxiko: TAploSustatiko;
begin
  result := fArxiko;         //     nil,       dk
end;

function TSuv8etoSustatiko.Teliko: TAploSustatiko;
{***************************************************************************************************
  Teliko    Perioxn.  Perioxn      
  Teliko,       dk.TrSus.
***************************************************************************************************}
var
  a: TAploSustatiko;
begin
  if fTeliko <> nil then result := fTeliko
  else begin
    a := fArxiko;
    while not (a is TTelosGrammns) do                      //     
      a := a.fEpomevo;                                                   //    dk
    result := (a as TTelosGrammns).dk.TrSus;
  end;
end;

function TSuv8etoSustatiko.Kwdikas: WideString;
{***************************************************************************************************
           .
        .
 :  fArxiko, fTeliko     .
      ,  fteliko=nil      .
***************************************************************************************************}
var
  a, tel: TAploSustatiko;
begin
  a := fArxiko;
  tel := a;
  result := '';
  while (a <> fTeliko) and (a <> nil) do         //thank you   for the bug report
  begin
    if a.fTupos in [sAmp, sSxolio] then                        //,   &  
    else if not (a.fTupos in [sKevo, sTelosGrammns]) then
      result := result + a.fKwdikas
    else if (Length(result) > 0) and (result[Length(result)] <> ' ') then
      if (fteliko = nil) and (a.fTupos = sTelosGrammns) then
        fteliko := a
      else
        result := result + ' ';                          //       
    tel := a;
    if fteliko <> a then              //  fteliko   nil,       a
      a := a.fEpomevo;
  end;
  if fTeliko = nil then
    fTeliko := tel
  else if fTeliko.fTupos <> sTelosGrammns then
    result := result + fTeliko.fKwdikas;
end;

{ TSta8era }

function TSta8era.DesTimn: Variant;
begin
  result := fTimn;
end;

function TSta8era.Leitourgies: Word;
begin
  result := inherited Leitourgies or lsDesTimn or lsSta8ernTimn or lsSta8ernApotimnsn;
end;

function TSta8era.TuposDedomevwv: TVarType;
begin
  case fTupos of
    sAkeraiaSt:       result := varInt64;
    sPragmatiknSt:    result := varDouble;
    sAlfari8mntiknSt: result := varOleStr;
  else
    result := varBoolean;
  end;
end;

procedure TSta8era.Apotimnse;
resourcestring
  sPoluMegalosAkeraios = '        .'
   + '         -9223372036854775808  9223372036854775807.';
  sPoluMegalosPragmatikos = '        .'
     + '        [+/-]  5.0*10^-324  1.7*10^308,'
     + '  15-16  ';
var
  i: int64;
  d: double;
  code: integer;
begin
  if fTupos = sAkeraiaSt then
  begin
{$IFOPT R+}{$DEFINE RestoreRangeChecking}{$ELSE}{$UNDEF RestoreRangeChecking}{$ENDIF}{$R-}
    Val(fKwdikas, i, code);            //todo:           {IF}
{$IFDEF RestoreRangeChecking}{$R+}{$ENDIF}
    if code = 0 then
      fTimn := VarAsType(i, varInt64)
    else
      raise ELektikoLa8os.Create(sPoluMegalosAkeraios, Perioxn);
  end else if fTupos = sPragmatiknSt then
  try
    Val(fKwdikas, d, code);                    //       {$R-}
    if code = 0 then
      fTimn := VarAsType(d, varDouble)
    else raise ELa8osEktelesns.Create;                         //      
  except
    raise ELektikoLa8os.Create(sPoluMegalosPragmatikos, Perioxn)
  end else if fTupos = sAlfari8mntiknSt then
  begin
    if fKwdikas[1] = '''' then fTimn := VarAsType(
      WideStringReplaceAll(Copy(fKwdikas, 2, Length(fKwdikas) - 2), '''''', ''''), varOleStr)
    else fTimn := VarAsType(
      WideStringReplaceAll(Copy(fKwdikas, 2, Length(fKwdikas) - 2), '""', '"'), varOleStr);
  end else                                                               //tupos = sAln8ns  sCeudns
    fTimn := VarAsType(fTupos = sAln8ns, varBoolean);
end;

{ TEkfrasnAvagvwristikou }

function TEkfrasnAvagvwristikou.DesTimn: Variant;
begin
  if avagvwristiko = nil then
    raise ESuvtaktikoLa8os.CreateFmt(sAgvwstoAvagvwristiko, [fKwdikas], Perioxn)
  else
    try
     result := avagvwristiko.timn;
    except
      on E: ELa8osEktelesns do     //         
        raise ELa8osEktelesns.Create(E.Message, Perioxn);
      else
        raise;
    end;
end;

procedure TEkfrasnAvagvwristikou.QeseTimn(const t: Variant);
begin
  if avagvwristiko = nil then
    raise ESuvtaktikoLa8os.CreateFmt(sAgvwstoAvagvwristiko, [fKwdikas], Perioxn)
  else avagvwristiko.timn := t;
end;

function TEkfrasnAvagvwristikou.Leitourgies: Word;
begin
  if avagvwristiko = nil then
    raise ESuvtaktikoLa8os.CreateFmt(sAgvwstoAvagvwristiko, [fKwdikas], Perioxn)
  else result := avagvwristiko.Leitourgies;
end;

function TEkfrasnAvagvwristikou.TuposDedomevwv: TVarType;
begin
  if avagvwristiko = nil then
    raise ESuvtaktikoLa8os.CreateFmt(sAgvwstoAvagvwristiko, [fKwdikas], Perioxn)
  else result := avagvwristiko.TuposDedomevwv;
end;

function TEkfrasnAvagvwristikou.Perigrafn: WideString;
begin
  if avagvwristiko = nil then
    result := ' '
  else result := avagvwristiko.Perigrafn;
end;

{ TAvagvwristiko }

class function TAvagvwristiko.SarwseEkfrasn(dk: TDiaxeiristnsKwdika): TSustatiko;
{***************************************************************************************************
   subclassed           
     ,       .
***************************************************************************************************}
begin
  result := dk.TrSus;
  dk.Epomevo;
end;

{ TListaSustatikwv }

function TListaSustatikwv.DesStoixeio(i: integer): TSustatiko;
begin
  result := sustatika[i];
end;

procedure TListaSustatikwv.QeseStoixeio(i: integer; s: TSustatiko);
begin
  sustatika[i] := s;
end;

function TListaSustatikwv.DesPln8os: integer;
begin
  result := Length(sustatika);
end;

procedure TListaSustatikwv.Eisagwgn(s: TSustatiko);
begin
  SetLength(sustatika, Length(sustatika) + 1);
  sustatika[High(sustatika)] := s;
end;

destructor TListaSustatikwv.Destroy;
var
  i: integer;
begin
  for i := Low(sustatika) to High(sustatika) do
    if not sustatika[i].InheritsFrom(TAploSustatiko) then     //   dk,   
      FreeAndNil(sustatika[i]);
  sustatika := nil;
  inherited;
end;

{ TDiaxeiristnsKwdika }

procedure TDiaxeiristnsKwdika.Apeleu8erwse;
{***************************************************************************************************
     ,  .
***************************************************************************************************}
var
  p, s: TSustatiko;
begin
  s := basnAplwv;
  while s <> nil do
  begin
    p := s;
    Assert(s is TAploSustatiko);
    s := (s as TAploSustatiko).fEpomevo;
    FreeAndNil(p);                                                //  TTelosGrammns
  end;
  basnAplwv := nil;
  telnGrammns := nil;
end;

function TDiaxeiristnsKwdika.EpomevoSnmavtiko: TAploSustatiko;
{***************************************************************************************************
        .       (&)
            .
  EpomevoSnmavtiko    ,    .
***************************************************************************************************}
var
  brnkaAmp: boolean;                          //  &,      
  metaToAmp: TAploSustatiko;
begin
  result := fTrSus;
  if result.fEpomevo <> nil then
    result := result.fEpomevo;
  repeat
    brnkaAmp := false;
    while result.fTupos in [sKevo, sSxolio] do
      result := result.fEpomevo;
    if result.fTupos = sTelosGrammns then
    begin
      metaToAmp := result.fEpomevo;
      while metaToAmp.fTupos = sKevo do      //    nil,    
        metaToAmp := metaToAmp.fEpomevo;                                            //sTelosGrammns.
      if metaToAmp.fTupos = sAmp then
      begin
        brnkaAmp := true;
        result := metaToAmp.fEpomevo;   //    nil,     sAmp
      end;
    end;
  until not brnkaAmp and not (result.fTupos in [sKevo, sSxolio]);
  la8osAmp := result.fTupos = sAmp;   //,           amp,
end;                                      //       ""  .

procedure TDiaxeiristnsKwdika.QeseTrSus(const trSus: TAploSustatiko);
begin
  fTrSus := trSus;
  if fTrSus is TEkfrasnAvagvwristikou then
    daa(fTrSus as TEkfrasnAvagvwristikou);
  fEpSus := EpomevoSnmavtiko;
end;

function TDiaxeiristnsKwdika.DesKwdika: WideString;
var
  a: TAploSustatiko;
  ts: TWideStrings;
begin
  ts := TWideStringList.Create;
  a := self.basnAplwv;
  while a <> nil do
  begin
    ts.Add(a.fKwdikas);
    a := a.fEpomevo;
  end;
  Result := GetTextStr(ts);
  FreeAndNil(ts);
end;

procedure TDiaxeiristnsKwdika.QeseKwdika(const k: WideString);
var
  la: TLektikosAvalutns;
  l: TLejn;
  tg: Integer;                                                              //  
begin
  Apeleu8erwse;
  basnAplwv := TTelosGrammns.Create(sTelosArxeiou, '', self);
  tg := 1;
  SetLength(telnGrammns, 1);
  telnGrammns[0] := basnAplwv;
  fTrSus := basnAplwv;
  la := TLektikosAvalutns.Create(k);
  repeat
    l := la.EpomevnLejn;
    if l.tupos = sAvagvwristiko then
      fTrSus.fEpomevo := TEkfrasnAvagvwristikou.Create(l.tupos, l.kwdikas)
    else if l.tupos in [sAkeraiaSt, sPragmatiknSt, sAlfari8mntiknSt, sAln8ns, sCeudns] then
      fTrSus.fEpomevo := TSta8era.Create(l.tupos, l.kwdikas)
    else if l.tupos in [sTelosArxeiou, sTelosGrammns] then
    begin
      fTrSus.fEpomevo := TTelosGrammns.Create(l.tupos, l.kwdikas, self);
      if tg > High(telnGrammns) then      //   10000      10 sec
        SetLength(telnGrammns, 100 + MulDiv(Length(telnGrammns), 150, 100));
      telnGrammns[tg] := fTrSus.fEpomevo as TTelosGrammns;
      Inc(tg);
    end
    else
      fTrSus.fEpomevo := TAploSustatiko.Create(l.tupos, l.kwdikas);
    fTrSus := fTrSus.fEpomevo;
  until l.tupos = sTelosArxeiou;
  SetLength(telnGrammns, tg);
  FreeAndNil(la);
  QeseTrSus(basnAplwv);
end;

procedure TDiaxeiristnsKwdika.FtiajeKeva(ap: TAploSustatiko; keva: integer);
{***************************************************************************************************
      ,       #keva    ap, 
  ap,    .    sKevo   ap  ,  
  .   keva  <= 0    sKevo  '',   sKevo  .
    ,       sKevo,     
  .
***************************************************************************************************}
var
  metaApoKevo: TAploSustatiko;                                   //     
begin
  if (ap = nil) or (ap.fEpomevo = nil) then       //       
    Exit;
  if ap.fEpomevo.Tupos = sKevo then               //        
    ap := ap.fEpomevo;                                                      //   
  if ap.Tupos = sKevo then
    if keva <= 0 then                                                //  ,  
     ap.fKwdikas := '' 
    else ap.fKwdikas := StringOfChar(' ', keva)
  else if keva > 0 then                         //         ap
  begin
    metaApoKevo := ap.fEpomevo;
    ap.fEpomevo := TAploSustatiko.Create(sKevo, StringOfChar(' ', keva));
    ap.fEpomevo.fEpomevo := metaApoKevo;
  end;
end;

constructor TDiaxeiristnsKwdika.Create(daa: TDiadikasiaAvalusnsAvagvwristikou);
begin
  self.daa := daa;
end;

function TDiaxeiristnsKwdika.Epomevo: TAploSustatiko;
{***************************************************************************************************
       fTrSus.   ,    
       .
***************************************************************************************************}
resourcestring
  sLa8osXaraktnrasSuvexisnsGrammns = '    (&)    '
    + '  ,     ';
  sAkurosXaraktnras = ' : %s';
  sAvoixtoAlfari8mntiko = '        %s';
begin
  fTrSus := fEpSus;
  if la8osAmp then
  begin
    la8osAmp := false;                                                       //  ...
    raise ESuvtaktikoLa8os.Create(sLa8osXaraktnrasSuvexisnsGrammns, fTrSus.Perioxn);
  end;
  if fTrSus.Tupos = sAkuro then
    raise ELektikoLa8os.CreateFmt(sAkurosXaraktnras, [fTrSus.fKwdikas], fTrSus.Perioxn)
  else if fTrSus.Tupos = sAvoixtnAlfari8mntiknSt then
    raise ELektikoLa8os.CreateFmt(sAvoixtoAlfari8mntiko, [fTrSus.fKwdikas], fTrSus.Perioxn)
  else if fTrSus is TEkfrasnAvagvwristikou then
    daa(fTrSus as TEkfrasnAvagvwristikou);
  result := fTrSus;
  fEpSus := EpomevoSnmavtiko;
end;

function TDiaxeiristnsKwdika.EpomevoYpoprogramma: boolean;
{***************************************************************************************************
              
    .
***************************************************************************************************}
var
  palioTr: TAploSustatiko;
begin
  palioTr := fTrSus;
  while not (fTrSus.Tupos in [sDiadikasia, sSuvartnsn, sTelosArxeiou]) do
    fTrSus := EpomevoSnmavtiko;                                         // fEpSus  
  result := fTrSus.Tupos <> sTelosArxeiou;
  if not result then fTrSus := palioTr                       //      
  else TrSus := fTrSus;                                      //      fEpSus
end;

function TDiaxeiristnsKwdika.BresGrammn(s: TSustatiko): integer;
{***************************************************************************************************
   index       s.Arxiko,    
 zero-based.       ,    ,  -1.
***************************************************************************************************}
var
  a: TAploSustatiko;
  i: integer;
begin
  a := s.Arxiko;
  while not (a is TTelosGrammns) do
    a := a.fEpomevo;
  for i := Low(telnGrammns) to High(telnGrammns) do
    if telnGrammns[i] = a then
    begin
      result := i - 1;
      Exit;
    end;
  result := -1;
end;

function TDiaxeiristnsKwdika.SAS(s: TSnmeio): TAploSustatiko;            //(Aplo) SustatikoApoSnmeio
{***************************************************************************************************
     ,        .
     .          , 
            .
***************************************************************************************************}
var
  tel, tr: integer;                                         // tel    
  telA, trA: TAploSustatiko;
begin
  telA := nil;                                                     //     warnings
  tel := 0;                                                        //     warnings
  Inc(s.Y);                     //    telnGrammns  s.Y     1
  if Length(telnGrammns) = 0 then                           //     ...
    result := nil
  else if s.Y <= Low(telnGrammns) then     //    ,    
    result := basnAplwv.fEpomevo
  else if s.Y >= High(telnGrammns) then  //     ,    enter
    result := telnGrammns[High(telnGrammns) - 1]        //    ...
  else begin
    trA := telnGrammns[s.Y - 1];             //    ,   repeat
    tr := 0;
    repeat                   //        s.X    
      if trA.Tupos <> sKevo then      //   ,       enter
      begin
        tel := tr;
        telA := trA;
      end;
      trA := trA.fEpomevo;
      tr := tr + Length(trA.fKwdikas);
    until (tr >= s.X) or (trA is TTelosGrammns);
    if (telA is TTelosGrammns)                     //     ,
     or not (trA.fTupos in [sKevo, sTelosGrammns, sTelosArxeiou]) then   //    
      result := trA                              // ,  ,     enter
    else if (trA.fTupos = sKevo)                 //    ,   :
     and (s.X - tel > tr + Length(trA.fEpomevo.fKwdikas) - s.X)  //     
     and not (trA.fEpomevo.fTupos in [sKevo, sTelosGrammns, sTelosArxeiou]) then   // 
      result := trA.fEpomevo                                             //   
    else result := telA;              //    (    enter)
  end;
end;

procedure TDiaxeiristnsKwdika.MorfopoinseKwdika;
{***************************************************************************************************
              .
        ,       .
   ,      .

 :
------------------
 *     .
 *    div, mod      .
 *  ,       .
 *  +, - 
   - :     ), , 
   -  : (, ,  ,  

  :
-----------------------
 *     .
 *   , .. _(234), [], *2
 *   
 *   ,    :
 *  : <-, <=, >=, <>, _, + (), - (), _&, _!


        :
----------------------------------------------------
 & =>  
 ( =>   ,        ( )
 ) ] =>  ,  
 * / ^ .. [ => 
 + - => ()  
 + - => () 
 < <- <=, <>, =, >, >= => 
 , : =>  ,  

      :
---------------------------------------------
 *        , .. @, %+#
 *    DL,   

 :
-------------
  __  __ 1_+_3 __ 20 ___ 5
  __ (123)
***************************************************************************************************}
type
  Katngoria = (kDesmeumevn, kSta8era, kAvagvwristiko, kSumbolo, kAkuro);
const
// EsoxesDL    nibbles,        ޻
//        ޻.  nibble $3     .
// $00 =    
// $01 =      
// $02 =   ,      
// $10 =      ,    
// $11 =      ,    
// $20 =   _,      
// $30 =     ,    
// $31 =     ,    
  EsoxesDL: array[sDLLow .. sDLHigh] of byte = (
   $00 {sDiv},             $00 {sMod},           $00 {sA_M},               $00 {sA_T},
   $00 {sAkeraia},         $00 {sAkeraies},      $00 {sAln8ns},            $11 {sAlliws},
   $11 {sAlliws_Av},       $01 {sAv},            $00 {sApo},               $31 {sArxn},
   $01 {sArxn_Epavalnyns}, $00 {sBnma}, $01 {sGia},$00 {sGraye},             $00 {sDiabase},
   $31 {sDiadikasia},      $00 {sE},             $00 {sEpavalabe},         $02 {sEpileje},
   $00 {sEF},              $00 {sH},             $00 {sHM},                $00 {sKai},
   $00 {sKalese},          $00 {sLOG},           $00 {sLogikes},           $00 {sLogikn},
   $00 {sMe},
   $00 {sMe_Bnma},         $31 {sMetablntes},    $00 {sMexri},             $10 {sMexris_Otou},
   $01 {sOso},             $00 {sOxi},           $11 {sPeriptwsn},         $00 {sPragmatikes},
   $00 {sPragmatikn},      $31 {sProgramma},     $31 {sSta8eres},          $00 {sSYNnm},
   $31 {sSuvartnsn},       $00 {sT_R},           $10 {sTelos_Av},          $30 {sTelos_Diadikasias},
   $10 {sTelos_Epavalnyns},$20 {sTelos_Epilogwv},$30 {sTelos_Programmatos},$30 {sTelos_Suvartnsns},
   $00 {sTote},            $00 {sXaraktnras},    $00 {sXaraktnres},        $00 {sCeudns});
// KevaTrm     ,      nibble $3  ޻
//         ,        
//   .
   KevaTrm: array[sTRMLow..sTRMHigh] of byte = (
   $31 {sAmp},          $30 {sArParev8esn},      $01 {sDeParev8esn},         $00 {sEpi},
   $30 {sMovadiaioSuv}, $11 {sDuadikoSuv},       $01 {sKomma},               $30 {sMovadiaioMeiov},
   $11 {sDuadikoPlnv},  $00 {sDiastnma},         $00 {sDia},                 $01 {sAvwKatwTeleia},
   $11 {sMikrotero},    $11 {sAva8esnTimnsANSI}, $11 {sMikroteroHIsoANSI},   $11 {sDiaforoANSI},
   $11 {sIsov},         $11 {sMegalutero},       $11 {sMegaluteroHIsoANSI},  $00 {sArAgkuln},
   $01 {sDeAgkuln},     $00 {sDuvamn},           $11 {sAva8esnTimnsUNICODE}, $11 {sDiaforoUNICODE},
   $11 {sMikroteroHIsoUNICODE},                  $11 {sMegaluteroHIsoUNICODE});
var
  esoxn: integer;
  esoxnSuv: integer;                                                 //    
  esoxnPlnv: integer;                                                //    
  esoxnMndev: boolean;                                               //    
  brnkaPeriptwsn: boolean;                      //         
  prwtoSxolio: boolean;
  kevaSxoliou: integer;                      //         
  a, p, ps, arxnGrammns: TAploSustatiko;                      //()  , 
begin
  if not ru8miseis[ruMorfopoinsnMetaglwttisn].timn then
    Exit;
  p := basnAplwv;
  ps := p;
  if (p <> nil) and (p.fEpomevo <> nil) then           //     
    a := p.fEpomevo
  else Exit;
  esoxn := 0;
  esoxnSuv := 0;
  esoxnPlnv := 0;
  esoxnMndev := false;
//todo: ,      while a.tupos = kevo do epomeno;

  brnkaPeriptwsn := false;
  prwtoSxolio := a.Tupos = sSxolio;
  arxnGrammns := p;
  while a <> nil do
  begin
    if not (p.Tupos in [sKevo, sTelosGrammns, sSxolio, sAmp]) then //todo: more
      ps := p;                                                          //ps =  
    if ru8miseis[ruStoixisnKwdika].timn then                   //      
      if a.Tupos in [sTRMLow..sTRMHigh] then
      begin
        if a.Tupos in sSuv + sMeiov then
          if ps.Tupos in [sAkeraiaSt..sAvagvwristiko, sDeParev8esn, sDeAgkuln] then
            if a.Tupos in sSuv then
              a.fTupos := sDuadikoSuv
            else a.fTupos := sDuadikoPlnv
          else
            if a.Tupos in sSuv then
              a.fTupos := sMovadiaioSuv
            else a.fTupos := sMovadiaioMeiov;

        case KevaTRM[a.Tupos] and $F0 of
          $00: FtiajeKeva(p, 0);
          $10: FtiajeKeva(p, 1);
        end;
        case KevaTRM[a.Tupos] and $0F of
          $00: FtiajeKeva(a, 0);
          $01: FtiajeKeva(a, 1);
        end;
  //todo: , ,   ...
      end
      else if a.Tupos in [sMod, sDiv] then
      begin
        FtiajeKeva(p, 1);
        FtiajeKeva(a, 1);
      end
      else if not (a.Tupos in [sTRMLow..sTRMHigh])
          and not (p.Tupos in [sTRMLow..sTRMHigh])
          and not (a.Tupos = sSxolio) then
         FtiajeKeva(p, 1);

    case a.Tupos of
      sDLLow .. sDLHigh:
      begin
        if ru8miseis[ruDesmeumevesLejeis].timn < 2 then
          a.fKwdikas := mPKTDL(a.Tupos);
        if a.Tupos = sPeriptwsn then
          brnkaPeriptwsn := true;
        if not brnkaPeriptwsn or (a.Tupos <> sAlliws) then
        begin                               //        
          case EsoxesDL[a.Tupos] and $0F of
            $01: Inc(esoxnSuv, 2);
            $02: Inc(esoxnSuv, 4);                                                 //  
          end;
          case EsoxesDL[a.Tupos] and $F0 of
            $10: Inc(esoxnPlnv, 2);
            $20: Inc(esoxnPlnv, 4);                                         //  _
            $30: esoxnMndev := true;
          end;
        end;
      end;
      sAva8esnTimnsANSI:
        if ru8miseis[ruAva8esnTimns].timn = 1 then
        begin
          a.fTupos := sAva8esnTimnsUNICODE;
          a.fKwdikas := pPKTS[a.Tupos];
        end;
      sMikroteroHIsoANSI:
        if ru8miseis[ruSugkritikoiTelestes].timn = 1 then
        begin
          a.fTupos := sMikroteroHIsoUNICODE;
          a.fKwdikas := pPKTS[a.Tupos];
        end;
      sDiaforoANSI:
        if ru8miseis[ruSugkritikoiTelestes].timn = 1 then
        begin
          a.fTupos := sDiaforoUNICODE;
          a.fKwdikas := pPKTS[a.Tupos];
        end;
      sMegaluteroHIsoANSI:
        if ru8miseis[ruSugkritikoiTelestes].timn = 1 then
        begin
          a.fTupos := sMegaluteroHIsoUNICODE;
          a.fKwdikas := pPKTS[a.Tupos];
        end;
      sAva8esnTimnsUNICODE:
        if ru8miseis[ruAva8esnTimns].timn = 0 then
        begin
          a.fTupos := sAva8esnTimnsANSI;
          a.fKwdikas := pPKTS[a.Tupos];
        end;
      sDiaforoUNICODE:
        if ru8miseis[ruSugkritikoiTelestes].timn = 0 then
        begin
          a.fTupos := sDiaforoANSI;
          a.fKwdikas := pPKTS[a.Tupos];
        end;
      sMikroteroHIsoUNICODE:
        if ru8miseis[ruSugkritikoiTelestes].timn = 0 then
        begin
          a.fTupos := sMikroteroHIsoANSI;
          a.fKwdikas := pPKTS[a.Tupos];
        end;
      sMegaluteroHIsoUNICODE:
        if ru8miseis[ruSugkritikoiTelestes].timn = 0 then
        begin
          a.fTupos := sMegaluteroHIsoANSI;
          a.fKwdikas := pPKTS[a.Tupos];
        end;
      sAvagvwristiko:
        if (a as TEkfrasnAvagvwristikou).avagvwristiko <> nil then
          a.fKwdikas := (a as TEkfrasnAvagvwristikou).avagvwristiko.ovoma.Kwdikas;
      sTelosGrammns, sTelosArxeiou:
        begin
          esoxn := esoxn - esoxnPlnv;
          if esoxnMndev then
            esoxn := 0;
          if ru8miseis[ruStoixisnKwdika].timn and not prwtoSxolio and
           ( (p.Tupos <> sSxolio) or (arxnGrammns.fEpomevo.Tupos <> sKevo)
           or (arxnGrammns.fEpomevo.fEpomevo <> p) ) then       //   | |
              if arxnGrammns.fEpomevo is TTelosGrammns then
                FtiajeKeva(arxnGrammns, 0)
              else FtiajeKeva(arxnGrammns, esoxn);
//todo:       |  !|,                    
          if ru8miseis[ruStoixisnSxoliwv].timn and not prwtoSxolio and (p.Tupos = sSxolio) then
          begin                                                             //  
            kevaSxoliou := ru8miseis[ruDejiPeri8wrio].timn - Length(p.fKwdikas);
            p := arxnGrammns;                                      //     
            if p.fEpomevo.Tupos <> sSxolio then        //       
              while p.fEpomevo.fEpomevo.Tupos <> sSxolio do
              begin
                Dec(kevaSxoliou, Length(p.fEpomevo.fKwdikas));
                p := p.fEpomevo;
              end;
            if not (p.fEpomevo.Tupos in [sKevo, sSxolio]) then
            begin
              p := p.fEpomevo;
              Dec(kevaSxoliou, Length(p.fKwdikas));
            end;
            FtiajeKeva(p, kevaSxoliou);                //      
          end;
          esoxn := esoxn + esoxnSuv;
          esoxnSuv := 0;
          esoxnPlnv := 0;
          esoxnMndev := false;
          brnkaPeriptwsn := false;
          if a.fEpomevo <> nil then
          begin
            prwtoSxolio := a.fEpomevo.Tupos = sSxolio;
          end;
          arxnGrammns := a;
        end;
    end;

    p := a;
    a := a.fEpomevo;
  end;
end;

destructor TDiaxeiristnsKwdika.Destroy;
begin
  Apeleu8erwse;
  inherited;
end;

end.
